/* gvp.c - Generic Variable Parsing */
/* Created: Tue 28 Mar 2023 10:50:07 PM EDT malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: gvp.c,v 1.5 2023/04/08 14:52:17 malakai Exp $ */

/* Copyright © 2023 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of Locolor - Last Outpost terminal color library
 *
 * Locolor is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * Locolor is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <glib.h>

#include "gvp.h"

/* local #defines */
#ifndef MAX_STRING_LENGTH
#define MAX_STRING_LENGTH 8192
#endif

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))
#endif

/* ---- structs and typedefs ---- */

/* ---- local variable declarations ---- */

/* ---- local function declarations ---- */
int gvp_parse(char *src, GList **argloc);
int gvp_argsep(char *src);
int gvp_parse_arglist(char *src,GList **argloc);
int gvp_parse_varname(char *src,GList **argloc);
GList *free_gvp_list(GList *l);
char *gvp_argno(GList *argl,int n);

/* ---- code starts here ---- */

/* gvp_parse() - Generic Variable Parse function.
 *
 * This code is meant to slot into a higher level parser like the Diku family
 * act() function.  It allows writing more verbose substitiution names along
 * with optional arguments, and getting those values back as a GList.
 *
 * Point it to a string in src beginning with a '$'.  If the $ and what comes
 * after it is a valid generic variable, returns the number of bytes that were
 * consumed out of src.  Caller must provide the address of a GList *list, and
 * the list pointer will be updated to a new GList containg a (varname)
 * followed by each (arg).  Caller should free that list with
 * free_gvp_list(list) when no longer needed.  
 *
 * Variable names have to be ascii isalnum() strings.  The optional arguments
 * can be UTF-8.  Args may be optionally quoted with single or double quotes,
 * parens, or brackets.  Args are separated by a comma, colon, or spaces.  The
 * entire variable ends with a space, or it can be wrapped in curly brackets
 * after the dollar sign.
 *
 * Example variable formats:
 *  '$var '
 *  '${var}'
 *  '$var(arg) '
 *  '${var(two args)}'
 *  '${var("one arg")}'
 *  '${var(arg1:arg2)}'
 *  '${var(,)}' <- (two empty arguments!)
 *
 */
int gvp_parse(char *src, GList **argloc) {

	char bracket = '\0';
	char *s = src;
	int got;

	if(!s) return(0);
	if(!argloc) return(0);

	if(*s++ != '$') return(0);

	if (*s == '{') {
		bracket = *s++;
	}

	/* get the variable name. */
	if( (got = gvp_parse_varname(s,argloc)) ) {
		s+=got;
	} else {
		goto PARSE_ERROR;
	}

	/* start of argument list? */
	if(*s == '(') {
		s++;
		s+=gvp_parse_arglist(s,argloc);
	
		if(!*s || *s != ')') {
			goto PARSE_ERROR;
		} else {
			s++;
		}
	}

	if(bracket == '{') {
		if(!*s || *s++ != '}') {
			goto PARSE_ERROR;
		}
	}

	return(s-src);

	PARSE_ERROR:
	g_list_free_full(*argloc,g_free); /* ditch the list. */
	*argloc = NULL;
	
	return(0);
}

/* Convienince function, calls g_list_free_full with the correct free function,
 * g_free(). */
GList *free_gvp_list(GList *l) {
	if(l) {
		g_list_free_full(l,g_free); /* ditch the list. */
	}
	return(NULL);
}

/* detect argumnent list seperator characters. */
int gvp_argsep(char *src) {
	char argseps[] = ":,;";
	char *s = src;
	if(*s == ')') {
		/* End of vars marker. */
		return(1);
	} else if(isspace(*s)) {
		/* multiple whitespace is consumed */
		while(*s && isspace(*s)) s++;
		return(s-src);
	} else if(strchr(argseps,*src)) {
		return(1);
	} 
	return(0);
}


/* parse the options from a list of them. */
int gvp_parse_arglist(char *src,GList **argloc) {

	char word[MAX_STRING_LENGTH];
	char *s, *w;
	s=src;
	w = word;
	char quotes[] = "\'\"`";
	char *quote;
	int firstchar = 1;
	int escaped =0;
	GList *junkargs=NULL;
	int gvpskip;
	char *eow = word + sizeof(word);

	/* escaped -> next char is not special -> not escaped */
	/* firstchar -> looking for firstchar quote -> quotemode*/
	/* quotemode -> looking for closequote or eof */
	/* !quotemode -> looking for seperator or ) or eof */

	while(*s) {
		if(escaped) {
			escaped = 0;
			*w++ = *s++;
		} else if (*s =='\\') {
			escaped = 1;
			s++;
		} else if(firstchar) {
			/* this could be another gvp variable, in which case it can be handled recursively. */
			if(*s=='$') {
				if( (gvpskip = gvp_parse(s,&junkargs)) ){
					w+=g_snprintf(w,MIN(gvpskip+1,eow-w),"%s",s);
					s+=gvpskip;
					*w='\0';
					junkargs = free_gvp_list(junkargs);
					continue;
				}
			}
			/* Or it could be a quoted string. */
			quote = strchr(quotes,*s);
			if(quote) {
				s++;
			} else if (!gvp_argsep(s) && (*s != ')')) {
				*w++ = *s++;
			} 
			firstchar = 0;
		} else if (quote) {
			/* see another of that first quote?  leave quotemode.*/
			if(*s == *quote) {
				quote = NULL;
				s++;
			} else {
				*w++ = *s++;
			}
		} else if (gvp_argsep(s) || (*s==')')) {
			/* found an argument seperator or an end of vars marker, either of
			 * which terminates the end of the argument.  Copy out the word
			 * into the list. */
			*w='\0';
			*argloc = g_list_append(*argloc,g_strdup(word));
			/* reset the word for the next one. */
			w=word;
			firstchar = 1;
			if(*s!=')') {
				/* Not done yet, so skip over the seperator. */
				s+=gvp_argsep(s);
			} else {
				/* it is the end of args operator, so we are done. */
				return(s-src);
			}
		} else {
			/* copy the character into the word. */
			*w++ = *s++;
		}
	}

	return(0);
}

/* parse out the arg0 variable name. */
int gvp_parse_varname(char *src,GList **argloc) {
	char *s = src;	
	char word[MAX_STRING_LENGTH];
	char *w, *eow;
	w = word;
	eow = word+sizeof(word);

	for(; *s && isalnum(*s) && (w<eow-1);) {
		*w++=*s++;
	}
	*w = '\0';
	*argloc = g_list_append(*argloc,g_strdup(word));
	return(s-src);
}

char *gvp_argno(GList *argl,int n) {
	return((char*)(g_list_nth_data(argl,n)));
}


/* --- test routines. --- */

/* sanity tests for different combinations. */
int gvp_test(void) {

	struct test_struct {
		char *input;
		int argc;
		int bytes;
	};

	struct test_struct test[] = {
		{"",0,0},
		{"$",0,0},
		{"$a",1,2},
		{"$a ",1,2},
		{"${a",0,0},
		{"${a}",1,4},
		{"${a} ",1,4},
		{"$(",0,0}, 
		{"$()",0,0}, 
		{"${()}",0,0}, 
		{"$a(",0,0}, 
		{"$a()",2,4},
		{"$a() ",2,4},
		{"${a()}",2,6},
		{"$a(b)",2,5},
		{"$a(,)",3,5},
		{"$a(b,c)",3,7},
		{"$a(b:c)",3,7},
		{"$a(b c)",3,7},
		{"$a(b    c)",3,10},
		{"$a(b c  d,e:f)",6,14},
		{"$a('b c  d',e:f)",4,16},
		{"$a(b,\"c c\",'d d',`e e`)",5,23},
		{"$a('You cant take \"that\" or `this` with you: nope.')",2,52},
		{"$a('You can\\'t take \"that\" or `this` with you: nope.')",2,54}, /*escaped quote */
		{"$v(ain't:ain't)",3,15}, /* quote must be first char of arg */
		{"$a('b",0,0},
		{"$a('b)",0,0},
		{"$a('b)',c)",3,10}, /* ) in quote doesn't end parsing. */
		{"$(Blue)red",0,0}, 
		{"$italic('This is $fg(Red)colorized text.$fg(DefaultFG)')",2,56}, 
		{"$bold(some \"$italic('This is $fg(Red)colorized text.$fg(DefaultFG))\" text)",4,74}, 
		{"$a($b($c($d('e'))))",2,19},
		{"$a($b(),$c(d))",3,14}
	};

	int failed = 0;
	int passed = 0;

	int t;
	for(t=0;t<ARRAY_SIZE(test);t++) {
		
		GList *args = NULL;
		int bytes = gvp_parse(test[t].input,&args);
		int argc = g_list_length(args);

		int bytescheck = (bytes == test[t].bytes);
		int argccheck = (argc == test[t].argc);
		int pass;

		pass= (bytescheck && argccheck);
		if( !pass ) failed ++;

		if( !pass || pass )  {
			fprintf(stdout,"Test %d:\n",t);
			fprintf(stdout,"\tinput: '%s'\n",test[t].input);
			fprintf(stdout,"\targc %d %s expected %d, bytes %d %s expected %d \n",
				argc,
				(argccheck)?"==":"!=",
				test[t].argc,
				bytes,
				(bytescheck)?"==":"!=",
				test[t].bytes
			);
			int count=0;
			for(GList *i=args;i;i=i->next) {
				fprintf(stdout,"\t%d: '%s'\n",count,(char *)(i->data));
				count++;
			}
		}

		if(!pass) {
			fprintf(stdout,"\tTest %d FAILED:\n",t);
			break;
		}

		g_list_free_full(args,g_free); /* ditch the list. */
	}

	if(failed) {	
		fprintf(stdout,"\n--- FAILED %d of %d Tests\n",failed,t);
		return(0);
	} 
	printf("\n--- PASSED %d Tests\n",t);
	return(1);
}

