/* loact.c - part of ANSI Terminal Color Handler */
/* Created: Wed 05 Apr 2023 10:27:29 PM EDT malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: loact.c,v 1.6 2023/05/16 13:42:19 malakai Exp $ */

/* Copyright © 2023 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of Locolor - Last Outpost terminal color library
 *
 * Locolor is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * Locolor is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <glib.h>

#include "gvp.h"
#include "locolor.h"
#include "loact.h"

/* ---- local #defines ---- */
#ifndef ARRAY_SIZE
#define ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))
#endif

#ifndef MAX_STRING_LENGTH
#define MAX_STRING_LENGTH 16384
#endif

/* ---- local structs and typedefs ---- */


/* ---- local variable declarations ---- */

/* a hash table to hold variable substitution functions for locolor_act() */
GHashTable *act_functions;

/* ---- local function declarations ---- */

/* local initialization routines */

/* builtin locolor_act functions. */
/* typedef int lo_actfn_t(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data); */
lo_actfn_t act_bg;
lo_actfn_t act_blend;
lo_actfn_t act_default;
lo_actfn_t act_fg;
lo_actfn_t act_fgbg;
lo_actfn_t act_nyan;
lo_actfn_t act_rainbow;
lo_actfn_t act_reset;
lo_actfn_t act_bold;
lo_actfn_t act_faint;
lo_actfn_t act_italic;
lo_actfn_t act_underline;
lo_actfn_t act_blink;
lo_actfn_t act_reverse;
lo_actfn_t act_crossedout;
lo_actfn_t act_doubleul;
lo_actfn_t act_rgba;
lo_actfn_t act_hsva;
lo_actfn_t act_style;

/* local other useful code. */
int act_generic_attr(GList *args, char *dst, size_t dlen, rgba_context *ctx, ansi_ca attr);
int act_blend_backend(GList *args, char *dst, size_t dlen, style_blendfn_t *fn,rgba_context *ctx, void *data);

/* ---- code starts here ---- */


/* Init the act extention hash table.  Call this first before using anything
 * else! */
void locolor_act_init(int use_builtins) {

	if (act_functions) {
		/* a reinit? */
		locolor_act_quit();
	}
	act_functions = g_hash_table_new_full(
		g_str_hash,     /* standard string hash function. */
		g_str_equal,    /* standard string compare function. */
		g_free,    /* standard string compare function. */
		NULL		/* No need to free, we stored a function pointer directly. */
	);

	if(use_builtins) {
		locolor_act_init_builtins();
	}
}

/* register an act variable substitution routine with the parser. */
int locolor_act_register(char *name,lo_actfn_t *fn) {
	return(g_hash_table_insert(act_functions,g_strdup(name),fn));
}

/* unregister an act variable from the parser. */
int locolor_act_unregister(char *name) {
	return(g_hash_table_remove(act_functions,name));
}

/* automatically add these builtin commands to the locolor_act parser */
void locolor_act_init_builtins(void) {

	locolor_act_register("bg",act_bg);
	locolor_act_register("default",act_default);
	locolor_act_register("fg",act_fg);
	locolor_act_register("fgbg",act_fgbg);
	locolor_act_register("nyan",act_nyan);
	locolor_act_register("reset",act_reset);
	locolor_act_register("rainbow",act_rainbow);
	locolor_act_register("rgba",act_rgba);
	locolor_act_register("hsva",act_hsva);
	locolor_act_register("style",act_style);
	locolor_act_register("blend",act_blend);

	locolor_act_register("italic",act_italic);
	locolor_act_register("bold",act_bold);
	locolor_act_register("faint",act_faint);
	locolor_act_register("underline",act_underline);
	locolor_act_register("blink",act_blink);
	locolor_act_register("reverse",act_reverse);
	locolor_act_register("crossedout",act_crossedout);
	locolor_act_register("doubleul",act_doubleul);

}


/* deallocates the act hash table.  Call on exit. */
void locolor_act_quit(void) {
	/* dump the whole act hash. */
	if(act_functions) g_hash_table_destroy(act_functions);
	act_functions = NULL;
}


/* example code for how one might parse out color directives and replace them
 * with locolor calls.  If you were using locolor in a game, you'd probably
 * incorporate some of these methods into your own act() function.  This
 * implemenation is mostly for testing using the demo code. */
int locolor_act(char *src, char *dst, size_t dlen, rgba_context *ctx) {

	GList *gvp_args = NULL;
	char *s;
	char *d,*eod;
	int varlen;
	char *varname;
	rgba_t color;
	char *arg;
	lo_actfn_t *actfn;
	int doneparsing = 0;

	ctx = resolve_ctx(ctx); 

	s = src;
	d = dst;
	eod = dst+dlen;

	for(;*s;s++) {
		if((*s == '$')) {
			if(*(s+1) == '-') {
				/* special symbol "$-" will cancel any remaining substitutions. */
				s+=2;
				break;  /* done parsing for vars. */
			} else if(*(s+1) == '$') {
				/* special symbol "$$" Is just a "$" */
				d+= g_snprintf(d,eod-d,"$");
				s++;
			} else if( (varlen=gvp_parse(s,&gvp_args)) ) { 
				/* see if its a "$gvp" parsable variable. */
				/* Is varname a registered variable function? */
				varname = (char *)(gvp_args->data);
				if(actfn = (lo_actfn_t *)g_hash_table_lookup(act_functions,varname)) {
					/* Before calling the function, recurse locolor_act over the arguments */
					for(GList* arg=gvp_args->next;arg;arg=arg->next) {
						/* Can't think of a reason to alloc more than the
						 * remaining destination space. */
						gchar *rarg = (char*)g_malloc(dlen);
						locolor_act((char*)(arg->data),rarg,dlen,ctx);
						g_free(arg->data);
						arg->data = rarg;
					}
					/* Now run it! */
					d+=((actfn)(gvp_args,d,eod-d,ctx,NULL));
					s+=varlen-1; 
				} else {
					/* hummm... not registered as one of ours? Leave it alone.*/
					d+= g_snprintf(d,eod-d,"$");
				}
				/* Don't forget to throw out the arguments we've parsed! */
				gvp_args = free_gvp_list(gvp_args);
			} else {
				/* ...Dunno what this '$' is.  Copy it to the dst. */
				if(d<eod-1) *d++ = *s;
			}
		} else {
			if(d<eod-1) *d++ = *s;
		}
	}
	/* if the parser was canceled with a $-, there'll still be bytes left over to copy.*/
	for(;*s && (d<eod-1);s++) *d++ = *s;
	*d = '\0';
	return(d-dst);
}

/* A simple wrapper around printf that will run locolor_act over the output.
 * Used by the testing and demo code to keep things clean. */
void locolor_printf(rgba_context *ctx, char *str, ...)
{

	va_list ap;
	char vbuf[MAX_STRING_LENGTH];
	char *v,*eov;
	char *src;

	v = vbuf;
	eov = vbuf + sizeof(vbuf)-1;

	va_start(ap,str);
	g_vsnprintf(v,eov-v,str,ap);
	va_end(ap);

	src = g_strdup(vbuf);

	v = vbuf;
	locolor_act(src,v,eov-v,ctx);
	g_free(src);
	fputs(vbuf,stdout);
}

/* Draw the text as a nyan cat banner.  Silly, but a good test. */
int act_nyan(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	
	char *message;
	char *d,*eod;
	rgba_t blue;
	GList *stl = NULL;
	style_t *style;

	d=dst;
	eod=dst+dlen;
	message = gvp_argno(args,1);

	stl = g_list_append(stl,new_style_parse("BrightWhite/Blue/",ctx));
	stl = g_list_append(stl,new_style_parse("BrightWhite/Blue/",ctx));

	d+= locolor_act("$blend(' ⠕ ⠖ ⠌ ⠓ ⡡ ','Grey3/DefaultBG/','BrightWhite/Blue/')", d,eod-d,ctx );
	d+= style_blend_text(d,eod-d,message,stl,style_blendfn_rainbow,ctx);
	d+= locolor_act(
		"$style(BrightWhite/Blue/)≋≋"
		"$fg(Grey12)╰$fgbg(BrightRed,#FFC0CB)⠝⠥$bg(Blue)$fg(Grey12)😽"
		"$blend(' ⠕ ⠖ ⠌ ⠓ ⡡ ','BrightWhite/Blue/','Grey3/DefaultBG/')"
		"${reset}",
		d,eod-d,ctx
	);

	g_list_free_full(stl,(GDestroyNotify)free_style);

	return(d-dst);

}

/* set the foreground */
int act_fg(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	
	char *d = dst;
	char *eod = dst+dlen;
	char *arg;
	rgba_t color;

	if((arg = gvp_argno(args,1))) {
		color = rgba_parse(arg,ctx);
	} else {
		color = ctx->default_style.fg;
	}

	ctx->fgbg_mode = ANSI_FG;
	d+=sgr_ansiColor(d,eod-d,color,ctx);

	return(d-dst);
}

/* set the background */
int act_bg(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	
	char *d = dst;
	char *eod = dst+dlen;
	char *arg;
	rgba_t color;

	if((arg = gvp_argno(args,1))) {
		color = rgba_parse(arg,ctx);
	} else {
		color = ctx->default_style.bg;
	}

	ctx->fgbg_mode = ANSI_BG;
	d+=sgr_ansiColor(d,eod-d,color,ctx);

	return(d-dst);
}

/* set both */
int act_fgbg(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	
	char *d = dst;
	char *eod = dst+dlen;
	char *arg;
	rgba_t color;

	if((arg = gvp_argno(args,2))&&*arg) {
		color = rgba_parse(arg,ctx);
	} else {
		color = ctx->default_style.bg;
	}
	ctx->fgbg_mode = ANSI_BG;
	d+=sgr_ansiColor(d,eod-d,color,ctx);

	if((arg = gvp_argno(args,1))&&*arg) {
		color = rgba_parse(arg,ctx);
	} else {
		color = ctx->default_style.fg;
	}
	ctx->fgbg_mode = ANSI_FG;
	d+=sgr_ansiColor(d,eod-d,color,ctx);

	return(d-dst);
}

/* set the style */
int act_style(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	
	char *d = dst;
	char *eod = dst+dlen;
	char *arg;
	style_t *style;

	style = style_dup(&ctx->default_style);

	if((arg = gvp_argno(args,1))) {
		if(!style_sscanf(arg,&style,ctx)) {
			if(*arg != '\0') {
				/* indicate not found */
				style->fg = 0xdeadbeef;
				style->bg = 0xdeadbea7;
				style->attr[ANSI_CA_CROSSEDOUT] = TRUE;
				style->attr[ANSI_CA_ITALIC] = TRUE;
			}
		}
	} else {
		d+=sgr_ansiNormal(d,eod-d,ctx);
	}

	d+=sgr_ansiStyle(d,eod-d,style,ctx);
	free_style(style);

	return(d-dst);
}

/* default alpha style blend. */
int act_blend(GList *args, char *dst, size_t dlen,rgba_context *ctx, void *data) {
	return(act_blend_backend(args,dst,dlen,NULL,ctx,data));
}

int act_rainbow(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_blend_backend(args,dst,dlen,style_blendfn_rainbow,ctx,data));
}


/* generic blender with a selectable blend function, meant to be called by
 * other act_blend type calls.   Text is in arg 1, styles are as many as you
 * want after that.  Currentstyle is restored when text is printed.  */
int act_blend_backend(GList *args, char *dst, size_t dlen, style_blendfn_t *fn,rgba_context *ctx, void *data) {
	
	char *d = dst;
	char *eod = dst+dlen;
	char *arg;
	style_t *style;
	GList *styles = NULL;
	GList *argp = NULL;
	style_t *currentstyle;

	currentstyle = style_dup(&ctx->current_style);

	/* build a list of style definitions starting with 2nd arg. */
	for(argp=g_list_nth(args,2);argp;argp=argp->next) {
		style = new_style();
		if(style_sscanf((char*)(argp->data),&style,ctx)) {
			styles = g_list_append(styles,style);
		} else {
			free_style(style);
		}
	}

	d+=style_blend_text(d,eod-d,gvp_argno(args,1),styles,fn,ctx);
	d+=sgr_ansiStyle(d,eod-d,currentstyle,ctx);

	g_list_free_full(styles,(GDestroyNotify)free_style);
	free_style(currentstyle);

	return(d-dst);
}

/* send ansiNormal */
int act_reset(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(sgr_ansiNormal(dst,dlen,ctx));
}

/* generic, attribute enable, send arg as message, restore attribute. */
int act_generic_attr_msg(GList *args, char *dst, size_t dlen, rgba_context *ctx, ansi_ca attr) {

	char *d = dst;
	char *eod = dst+dlen;
	char *message = gvp_argno(args,1);
	gboolean state = (ctx->current_style.attr[attr])?1:0;

	if(message && (*message != '\0')) {
		d+=sgr_ansiAttr(d,eod-d,attr,TRUE,ctx);
		d+=g_snprintf(d,eod-d,"%s",message);
		d+=sgr_ansiAttr(d,eod-d,attr,state,ctx);
	} 
	return(d-dst);
}

/* generic act parser for one-shot attributes.  Without an argument, or with
 * empty argument, will toggle the attribute.  If attribute is {true,on,1} then
 * enable the attribue, otherwise disable it.  So for example, $italic or
 * $italic() will toggle the italic state.  $italic(on) will turn the attribute
 * on.
 */
int act_generic_attr(GList *args, char *dst, size_t dlen, rgba_context *ctx, ansi_ca attr) {

	char *d = dst;
	char *eod = dst+dlen;
	char *message = gvp_argno(args,1);
	gboolean state = (ctx->current_style.attr[attr])?1:0;

	if(message && (*message != '\0')) {
		if( !strcasecmp(message,"true") ||
			!strcasecmp(message,"on") ||
			!strcasecmp(message,"1")
		) {
			d+=sgr_ansiAttr(d,eod-d,attr,TRUE,ctx);
		} else {
			d+=sgr_ansiAttr(d,eod-d,attr,FALSE,ctx);
		}
	} else {
		/* no arg, or empty arg, toggle. */
		d+=sgr_ansiAttr(d,eod-d,attr,state^1,ctx);
	}
	return(d-dst);
}

/* act for ANSI_CA_BOLD */
int act_bold(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_BOLD));
}

/* act for ANSI_CA_FAINT */
int act_faint(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_FAINT));
}

/* act for ANSI_CA_ITALIC */
int act_italic(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_ITALIC));
}

/* act for ANSI_CA_UNDERLINE */
int act_underline(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_UNDERLINE));
}

/* act for ANSI_CA_BLINK */
int act_blink(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_BLINK));
}

/* act for ANSI_CA_REVERSE */
int act_reverse(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_REVERSE));
}

/* act for ANSI_CA_CROSSEDOUT */
int act_crossedout(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_CROSSEDOUT));
}

/* act for ANSI_CA_DOUBLEUL */
int act_doubleul(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(act_generic_attr(args,dst,dlen,ctx,ANSI_CA_DOUBLEUL));
}

int act_default(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {
	return(sgr_ctxDefaults(dst,dlen,ctx));
}

/* return the color string as #RRGGBBAA */
int act_rgba(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {

	char *d = dst;
	char *eod = dst+dlen;
	char *name;
	rgba_t color;

	if((name = gvp_argno(args,1))) {
		color = rgba_parse(name,ctx);
	} else {
		color = 0xDEADBEEF;
	}
	d+=g_snprintf(d,eod-d,"#%08X",color);
	return(d-dst);
}

/* return the color string as #RRGGBBAA */
int act_hsva(GList *args, char *dst, size_t dlen, rgba_context *ctx, void *data) {

	char *d = dst;
	char *eod = dst+dlen;
	char *name;
	hsva_t color;

	if((name = gvp_argno(args,1))) {
		color.h = atof(name);
	} 
	if((name = gvp_argno(args,2))) {
		color.s = atof(name);
	} 
	if((name = gvp_argno(args,3))) {
		color.v = atof(name);
	} 
	if((name = gvp_argno(args,4))) {
		color.a = atof(name);
	} 

	d+=g_snprintf(d,eod-d,"#%08X",rgba_from_hsva(&color));
	return(d-dst);
}
