/* locolor.c - ANSI Terminal Color Handler */
/* Created: Tue 07 Mar 2023 10:34:32 PM EST malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: locolor.c,v 1.17 2023/05/16 13:42:19 malakai Exp $ */

/* Copyright © 2023 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of Locolor - Last Outpost terminal color library
 *
 * Locolor is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * Locolor is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <glib.h>

#include "gvp.h"
#include "locolor.h"

/* ---- local #defines ---- */
#define RED_CHANNEL(c)	((channel_t)(((uint32_t)(c) & 0xFF000000)>>24))
#define GREEN_CHANNEL(c)	((channel_t)(((uint32_t)(c) & 0x00FF0000)>>16))
#define BLUE_CHANNEL(c)	((channel_t)(((uint32_t)(c) & 0x0000FF00)>>8))
#define ALPHA_CHANNEL(c)	((channel_t)(((uint32_t)(c) & 0x000000FF)>>0))

#define RED_SHIFT(c)	(((uint32_t)(c) & 0xFF)<<24)
#define GREEN_SHIFT(c)	(((uint32_t)(c) & 0xFF)<<16)
#define BLUE_SHIFT(c)	(((uint32_t)(c) & 0xFF)<<8)
#define ALPHA_SHIFT(c)	(((uint32_t)(c) & 0xFF)<<0)

#define NAMED_COLOR_LEN	64
#define NAMED_STYLE_LEN	512

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))
#endif

#ifndef MAX_STRING_LENGTH
#define MAX_STRING_LENGTH 16384
#endif

/* ---- local structs and typedefs ---- */


/* ---- local variable declarations ---- */

colorchip_t ansi256[] = {
	/* The ansi std colors in the 256 palette. */
	[0] = { "Black", "#000", 0x000000ff, 0 }, { "Red", "#a00", 0xaa0000FF, 1 },
	{ "Green", "#0a0", 0x00aa00ff, 2 }, { "Yellow", "#aa0", 0xaaaa00FF, 3 },
	{ "Blue", "#00a", 0x0000aaff, 4 },  { "Magenta", "#a0a",0xaa00aaFF, 5 },
	{ "Cyan", "#0aa", 0x00aaaaff, 6 },  { "White", "#aaa", 0xaaaaaaFF, 7 },
	/* The ansi bright colors in the 256 palette. */
	[8] = { "BrightBlack", "#555", 0x555555ff, 8 },  { "BrightRed", "#f55", 0xFF5555FF, 9 },
	{ "BrightGreen", "#5f5", 0x55FF55ff, 10 }, { "BrightYellow", "#ff5", 0xFFFF55FF, 11 },
	{ "BrightBlue", "#55f", 0x5555FFff, 12 },  { "BrightMagenta", "#f5f",0xFF55FFFF, 13 },
	{ "BrightCyan", "#5ff", 0x55FFFFff, 14 },  { "BrightWhite", "#fff", 0xFFFFFFFF, 15 },
	/* The six cube colors run from [16] - [231], and will be filled in
	 * algorithmically by locolor_init() */
	/* The ansi256 grey colors start at 232. */
	[232] = { "Grey0", "#080808", 0x080808ff, 232 },  { "Grey1", "#121212", 0x121212ff, 233 },
	{ "Grey2", "#1c1c1c", 0x1c1c1cff, 234 },  { "Grey3", "#262626", 0x262626ff, 235 },
	{ "Grey4", "#303030", 0x303030ff, 236 },  { "Grey5", "#3a3a3a", 0x3a3a3aff, 237 },
	{ "Grey6", "#444444", 0x444444ff, 238 },  { "Grey7", "#4e4e4e", 0x4e4e4eff, 239 },
	{ "Grey8", "#585858", 0x585858ff, 240 },  { "Grey9", "#626262", 0x626262ff, 241 },
	{ "Grey10", "#6c6c6c", 0x6c6c6cff, 242 }, { "Grey11", "#767676", 0x767676ff, 243 },
	{ "Grey12", "#808080", 0x808080ff, 244 }, { "Grey13", "#8a8a8a", 0x8a8a8aff, 245 },
	{ "Grey14", "#949494", 0x949494ff, 246 }, { "Grey15", "#9e9e9e", 0x9e9e9eff, 247 },
	{ "Grey16", "#a8a8a8", 0xa8a8a8ff, 248 }, { "Grey17", "#b2b2b2", 0xb2b2b2ff, 249 },
	{ "Grey18", "#bcbcbc", 0xbcbcbcff, 250 }, { "Grey19", "#c6c6c6", 0xc6c6c6ff, 251 },
	{ "Grey20", "#d0d0d0", 0xd0d0d0ff, 252 }, { "Grey21", "#dadada", 0xdadadaff, 253 },
	{ "Grey22", "#e4e4e4", 0xe4e4e4ff, 254 }, { "Grey23", "#eeeeee", 0xeeeeeeff, 255 }
};

/* name, CSI Ps m Character Attributes (SGR) Ps for disabling,enabling. */
sgr_ca_t ansi_attributes[ANSI_CA_MAX] = {
	[ANSI_CA_NORMAL]=		{"Normal",		{"0","0"}},
	[ANSI_CA_BOLD]=			{"Bold",		{"22","1"}},
	[ANSI_CA_FAINT]=		{"Faint",		{"22","2"}},
	[ANSI_CA_ITALIC]=		{"Italic",		{"23","3"}},
	[ANSI_CA_UNDERLINE]=	{"Underline",	{"24","4"}},
	[ANSI_CA_BLINK]=		{"Blink",		{"25","5"}},
	[ANSI_CA_FASTBLINK]=	{"FastBlink",	{"25","6"}},
	[ANSI_CA_REVERSE]=		{"Reverse",		{"27","7"}},
	[ANSI_CA_HIDDEN]=		{"Hidden",		{"28","8"}},
	[ANSI_CA_CROSSEDOUT]=	{"CrossedOut",	{"29","9"}},
	[ANSI_CA_DOUBLEUL]=		{"DoubleUL",	{"24","21"}},
	[ANSI_CA_FRAMED]=		{"Framed",		{"54","51"}},
	[ANSI_CA_ENCIRCLED]=	{"Encircled",	{"54","52"}},
	[ANSI_CA_OVERLINED]=	{"Overline",	{"55","53"}},
	[ANSI_CA_SUPERSCRIPT]=	{"Superscript",	{"75","73"}},
	[ANSI_CA_SUBSCRIPT]=	{"Subscript",	{"75","74"}}
};

/* a (locally) global set of default system options, including the system
 * palette, initialized by locolor_init(); */
rgba_context *default_ctx = NULL;

/* Useful weights for different grescale Luma transforms. */
rgba_t luma_Linear = 0x555555ff;
rgba_t luma_ATSC = 0x36b612ff;
rgba_t luma_NTSC = 0x4c961dff;
rgba_t luma_HDR = 0x43ad0fff;

/* The default style */
style_t default_style_entry = {
	.name="Default",
	.definition="White/Black/Normal",
	.fg=0xaaaaaaFF,  /* ansi White */
	.bg=0x000000FF,  /* ansi Black */
	.attr[ANSI_CA_NORMAL]=TRUE,
	.opt[STYLE_FG]=TRUE,
	.opt[STYLE_BG]=TRUE,
	.opt[STYLE_ATTR]=TRUE
};


/* ---- function declarations ---- */

/* initialization routines */
void locolor_init(void);
void locolor_quit(void);
/* ... */
static void init_ansi216(void);
static void init_default_ctx(void);
static void quit_ansi216(void);
static void quit_default_ctx(void);

/* context manipulation */
ansi_fgbg color_mode_from_env(void);
rgba_context *new_rgba_context(void);
rgba_context *resolve_ctx(rgba_context *ctx);
void free_rgba_context(rgba_context *f);

/* type conversion */
rgba_t rgba_from_channel(channel_t r, channel_t g, channel_t b, channel_t a);
rgba_t rgba_from_double(double r, double g, double b, double a);
rgba_t rgba_from_drgba(drgba_t *d);
rgba_t rgba_from_hsva(hsva_t *h);
void rgba_to_ansi216idx(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a);
void rgba_to_channel(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a);
void rgba_to_double(rgba_t color, double *r, double *g, double *b, double *a);
void rgba_to_drgba(rgba_t color, drgba_t *d);
void rgba_to_hsva(rgba_t color, hsva_t *h);
/* ... */
static double hsvaltf(double n, hsva_t *h);

/* color manipulation functions */
double rgba_contrast_ratio(rgba_t A, rgba_t B);
double rgba_relative_luminance(rgba_t color);
double unit_clip(double c);
int rgba_distance2(rgba_t c1, rgba_t c2);
rgba_t rgba_alphablend(rgba_t A, rgba_t B);
rgba_t rgba_alpha(rgba_t c,channel_t a);
rgba_t rgba_ansi216(rgba_t color);
rgba_t rgba_contrasting_color(rgba_t fg, rgba_t bg, double mincr);
rgba_t rgba_from_ansi256(int idx);
rgba_t rgba_gammablend(rgba_t A, rgba_t B, double y);
rgba_t rgba_gamma(rgba_t c,double y);
rgba_t rgba_greyscale(rgba_t color);
rgba_t rgba_hsvblend(rgba_t A, rgba_t B, double amount, int ccw);
rgba_t rgba_hsvblend_long(rgba_t A, rgba_t B, double amount);
rgba_t rgba_hsvblend_path(rgba_t A, rgba_t B, double amount, int longway);
rgba_t rgba_hsvblend_short(rgba_t A, rgba_t B, double amount);
rgba_t rgba_invert(rgba_t a);
rgba_t rgba_multiply(rgba_t a, rgba_t b);
rgba_t rgba_opaque(rgba_t c);
rgba_t rgba_posterize(rgba_t color, double N);
rgba_t rgba_screen(rgba_t a, rgba_t b);
/* ... */
static double posterize_double(double c,double N);
static double drgba_distance2(drgba_t *c1, drgba_t *c2);


/* ANSI Terminal code functions. */
int sgr_ansi216(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ansiAttr(char *dst, size_t dlen, ansi_ca attr, gboolean state, rgba_context *ctx);
int sgr_ansiColor(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ansiNormal(char *buf, size_t size, rgba_context *ctx);
int sgr_ansiStyle(char *dst, size_t dlen, style_t *style, rgba_context *ctx);
int sgr_ansiTruecolor(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ctxDefaults(char *buf, size_t size, rgba_context *ctx);
/* ... */
static const char *ESC(rgba_context *ctx);
static int ansi216_index(channel_t ch);
static int is_current_color(rgba_t color,rgba_context *ctx);
static rgba_t set_current_color(rgba_t color,rgba_context *ctx);
static void printansi216(void);

/* colorchip related functions */
int lookup_color_name(char *name, rgba_context *ctx, rgba_t *color);
int load_colorchip_hash(char *filename, rgba_context *ctx);
int rgba_sscanf(char *buf,rgba_t *c, rgba_context *ctx);
int colorchip_sscanf(char *buf,colorchip_t *c, rgba_context *ctx);
colorchip_t *new_colorchip(void);
colorchip_t *new_colorchip_parse(char *name,rgba_context *ctx);
colorchip_t *rgba_closest_match(rgba_t color, colorchip_t *pal, int from, int to);
colorchip_hash_t *new_palette(void);
rgba_t rgba_parse(char *name,rgba_context *ctx);
void free_colorchip(colorchip_t *f);
void free_colorchip_hash(colorchip_hash_t *f);

/* style related functions */
int load_style_hash(char *filename, rgba_context *ctx);
int style_sscanf(char *buf, style_t **styleloc, rgba_context *ctx);
style_hash_t *new_style_hash(void);
style_t *lookup_style_name(char *name, rgba_context *ctx);
style_t *new_style(void);
style_t *new_style_parse(char *name,rgba_context *ctx);
style_t *style_cpy(style_t *d, style_t *s);
style_t *style_dup(style_t *s);
void free_style(style_t *f);
void free_style_hash(style_hash_t *f);

/* fancy style blending and formating functions */
int style_blend_text(char *dst,size_t dlen,char *src, GList *styles, style_blendfn_t *blendfn, rgba_context *ctx);
style_t *style_blendfn_list(double amount,GList *slist, style_blendfn_t *blendfn,rgba_context *ctx);
/* ... these are style_blendfn_t function prototypes. */
style_t *style_blendfn_generic( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_gamma( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_hsva( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_rainbow( style_t *a, style_t *b, double amount, rgba_context *ctx);
/* ... */

/* local other useful code. */

/* ---- code starts here ---- */

/* initialization routines */

/* Call this first before using anything else! */
void locolor_init(void) {
	init_ansi216();
	init_default_ctx();
}

/* and call this last. */
void locolor_quit(void) {
	quit_default_ctx();
	quit_ansi216();
}

/* calculates and assigns names to the middle of the palette. */
void init_ansi216(void) {
	int r,g,b;
	int index;
	rgba_t color;
	channel_t bucket[6] = { 0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff };
	char name[NAMED_COLOR_LEN];

	for(r=0;r<=5;r++) {
		for(g=0;g<=5;g++) {
			for(b=0;b<=5;b++) {
				index = 16 + (36 * r) + (6 * g) + (b);
				sprintf(name,"six%d%d%d", r,g,b);
				ansi256[index].name = g_strdup(name);
				color = rgba_from_channel(bucket[r],bucket[g],bucket[b],0xff);
				sprintf(name,"#%08X", color);
				ansi256[index].definition = g_strdup(name);
				ansi256[index].color = color;
				ansi256[index].index = index;
			}
		}
	}
}

/* create the default system options. */
void init_default_ctx(void) {

	colorchip_t *entry;
	style_t *st;

	if(default_ctx) {
		/* a re-init perhaps? */
		free_rgba_context(default_ctx);
	}

	default_ctx = new_rgba_context();

	/* copy the array of colorchip_t's from ansi256 into the palette lookup hash. */
	for(int i=0;i<ARRAY_SIZE(ansi256);i++) {
		entry = new_colorchip();
		entry->name = g_strdup(ansi256[i].name);
		entry->color = ansi256[i].color;
		entry->index = ansi256[i].index;
		g_hash_table_insert ( default_ctx->palette, g_strdup(entry->name), entry );
	}

	/* copy default into the style map. */
	st = style_dup(&default_style_entry);
	g_hash_table_insert(default_ctx->styles,g_strdup(st->name),st);

	return;
}

/* Frees up the names that were assigned to the middle of the palette. */
void quit_ansi216(void) {
	int r,g,b;
	int index;

	for(r=0;r<=5;r++) {
		for(g=0;g<=5;g++) {
			for(b=0;b<=5;b++) {
				index = 16 + (36 * r) + (6 * g) + (b);
				g_free(ansi256[index].name);
			}
		}
	}
}

/* frees it up. */
void quit_default_ctx(void) {
	if(default_ctx) {
		free_rgba_context(default_ctx);
	}
	default_ctx = NULL;
	return;
}

/* context manipulation */

/* look for environment variables that indicate a color terminal. */
ansi_fgbg color_mode_from_env(void) {
	if(getenv("COLORTERM")) {
		if(!strcasecmp(getenv("COLORTERM"),"truecolor")) {
			return(ANSI_TRUECOLOR);
		}
		return(ANSI_240);
	}
	/* you could retrun ANSI_MONO here too. */
	return(ANSI_240);
}

/* create a new rgba_context structure with reasonable defaults. */
rgba_context *new_rgba_context(void) {
	rgba_context *n;

	n = (rgba_context *)malloc(sizeof(rgba_context));
	n->gamma = 2.2;
	n->contrast_ratio=1.0;  /* 1.0 means do nothing. */
	n->palette = new_palette();
	n->system_palette_fallback = 1;

	n->styles = new_style_hash();
	n->system_styles_fallback = 1;

	n->default_style = default_style_entry;
	//n->color_mode = ANSI_MONO;  /* off by default. */
	n->color_mode = ANSI_240;   /* most compatible by default. */
	//n->color_mode = ANSI_TRUECOLOR; /* most colors by default. */
	n->fgbg_mode = ANSI_FG;
	n->unmodified_fg = n->default_style.fg;
	n->current_style = default_style_entry;
	n->current_style.fg = ~(n->default_style.fg); /* not likely to be true. */
	n->current_style.bg = ~(n->default_style.bg); /* not likely to be true. */

	n->debug = 0;
	return(n);
}

/* call this on any ctx handed to a function.  Makes it so that a NULL ctx is
 * replaced with the default_ctx. */
rgba_context *resolve_ctx(rgba_context *ctx) {
	if(ctx) return (ctx);
	if(!default_ctx) locolor_init();
	return(default_ctx);
}

/* does what it claims to. */
void free_rgba_context(rgba_context *f) {
	if(!f) return;
	if(f->palette) free_colorchip_hash(f->palette);
	if(f->styles) free_style_hash(f->styles);
	free(f);
}

/* type conversion */

/* given four channel_t's with a range of 0x00-0xFF, compose into a singgle rgba_t. */
rgba_t rgba_from_channel(channel_t r, channel_t g, channel_t b, channel_t a) {
	rgba_t c = 0;
	c |= RED_SHIFT(r);
	c |= GREEN_SHIFT(g);
	c |= BLUE_SHIFT(b);
	c |= ALPHA_SHIFT(a);
	return(c);
}

/* given four doubles with a range of 0.0-1.0, compose into a singgle rgba_t. */
rgba_t rgba_from_double(double r, double g, double b, double a) {
	rgba_t c = 0;

	c |= RED_SHIFT(round(unit_clip(r)*0xFF));
	c |= GREEN_SHIFT(round(unit_clip(g)*0xFF));
	c |= BLUE_SHIFT(round(unit_clip(b)*0xFF));
	c |= ALPHA_SHIFT(round(unit_clip(a)*0xFF));
	return(c);
}

/* return packed rgba_t from pointer to drgba_t struct */
rgba_t rgba_from_drgba(drgba_t *d) {
	rgba_t c = 0;

	if(!d) return (0xDEADBEEF);

	c |= RED_SHIFT(round(unit_clip(d->r)*0xFF));
	c |= GREEN_SHIFT(round(unit_clip(d->g)*0xFF));
	c |= BLUE_SHIFT(round(unit_clip(d->b)*0xFF));
	c |= ALPHA_SHIFT(round(unit_clip(d->a)*0xFF));

	return(c);
}

/* https://en.wikipedia.org/wiki/HSL_and_HSV */
rgba_t rgba_from_hsva(hsva_t *h) {

	if(!h) return (0xDEADBEEF);

	drgba_t d;
	d.r = hsvaltf(5.0,h),
	d.g = hsvaltf(3.0,h),
	d.b = hsvaltf(1.0,h),
	d.a = h->a;
	return( rgba_from_drgba(&d) );
}

/* decompose rgba_t color into the ansi216 index 0-5 */
void rgba_to_ansi216idx(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a) {
	channel_t ri,gi,bi,ai;

	rgba_to_channel(color,&ri,&gi,&bi,&ai);
	if(r) *r = ansi216_index(ri);
	if(g) *g = ansi216_index(gi);
	if(b) *b = ansi216_index(bi);
	if(a) *a = ansi216_index(ai);
}

/* decompose rgba_t color into the locations of four channel_t's. */
void rgba_to_channel(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a) {
	if(r) *r = RED_CHANNEL(color);
	if(g) *g = GREEN_CHANNEL(color);
	if(b) *b = BLUE_CHANNEL(color);
	if(a) *a = ALPHA_CHANNEL(color);
}

/* given an rgba_t color, decompose it into the locations of four doubles with
 * range of 0.0-1.0. */
void rgba_to_double(rgba_t color, double *r, double *g, double *b, double *a) {

	if(r) *r = (double)(RED_CHANNEL(color))/(0xff);
	if(g) *g = (double)(GREEN_CHANNEL(color))/(0xff);
	if(b) *b = (double)(BLUE_CHANNEL(color))/(0xff);
	if(a) *a = (double)(ALPHA_CHANNEL(color))/(0xff);

}

/* given an rgba_t color, decompose it into a drgba_t where members are doubles
 * with range of 0.0-1.0. */
void rgba_to_drgba(rgba_t color, drgba_t *d) {
	if(d) {
		d->r = (double)(RED_CHANNEL(color))/(0xff);
		d->g = (double)(GREEN_CHANNEL(color))/(0xff);
		d->b = (double)(BLUE_CHANNEL(color))/(0xff);
		d->a = (double)(ALPHA_CHANNEL(color))/(0xff);
	}

}



/* https://en.wikipedia.org/wiki/HSL_and_HSV */
void rgba_to_hsva(rgba_t color, hsva_t *h) {
	
	drgba_t d;
	double Xmax;
	double Xmin;
	double H; /* Hue */
	double V; /* value */
	double C; /* chroma */
	// double L; /* lightness */
	double Sv; /* saturation */
	
	rgba_to_drgba(color, &d);
	Xmax = MAX(d.r,MAX(d.g,d.b));
	V = Xmax;
	Xmin = MIN(d.r,MIN(d.g,d.b));
	C =  Xmax - Xmin;
	/* could add lightness to the hsva_t sometime.  Here's how to calcluate it.
	L = (Xmax + Xmin) / 2.0;
	*/
	H = 0.0;
	if(C != 0.0) {
		if(V == d.r) {
			H = 60 * ( fmod(((d.g-d.b)/C),6.0));
		} else if(V == d.g) {
			H = 60 * ( ((d.b-d.r)/C) + 2.0);
		} else if(V == d.b) {
			H = 60 * ( ((d.r-d.g)/C) + 4.0);
		}
	}
	if(V == 0.0) {
		Sv = 0.0;
	} else {
		Sv = C/V;
	}
	
	h->h=H;
	h->s=Sv;
	h->v=V;
	h->a=1.0;

	return;

}

/* https://en.wikipedia.org/wiki/HSL_and_HSV */
double hsvaltf(double n, hsva_t *h) {
	
	double k;
	double f;

	k = fmod( (n+((h->h)/60.0)) , 6.0);
	f = (h->v) - ((h->v)*(h->s)*(unit_clip(MIN(k,4.0-k))));
	return(f);

}

/* color manipulation functions */

/* https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef Can range
 * from 1 to 21 i.e. 1:1 to 21:1  */
double rgba_contrast_ratio(rgba_t A, rgba_t B) {

	double Al,Bl;
	double L1,L2;

	Al = rgba_relative_luminance(A);
	Bl = rgba_relative_luminance(B);
	
	L1 = MAX(Al,Bl);
	L2 = MIN(Al,Bl);

	return( (L1+0.05) / (L2+0.05) );
}

/* https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef */
double rgba_relative_luminance(rgba_t color) {
	drgba_t d;
	double R,G,B,L;

	rgba_to_drgba(color, &d);
	R = (d.r<=0.03928)?(d.r/12.0):(pow(((d.r+0.055)/1.055),2.4));
	G = (d.g<=0.03928)?(d.g/12.0):(pow(((d.g+0.055)/1.055),2.4));
	B = (d.b<=0.03928)?(d.b/12.0):(pow(((d.b+0.055)/1.055),2.4));
	L = (0.2126 * R) + (0.7152 * G) + (0.0722 * B); /* this is luma_ATSC fwiw. */
	return(L);
}

/* keep a double between [0.0-1.0] */
double unit_clip(double c) {
	if(c > 1.0) return(1.0);
	if(c < 0.0 ) return(0.0);
	return(c);
}

/* return the euclidian squared distance of two rgba colors. Alpha ignored.*/
int rgba_distance2(rgba_t c1, rgba_t c2) {
	int ds;

	ds= (RED_CHANNEL(c1)-RED_CHANNEL(c2)) * (RED_CHANNEL(c1)-RED_CHANNEL(c2)) +
		(GREEN_CHANNEL(c1)-GREEN_CHANNEL(c2)) * (GREEN_CHANNEL(c1)-GREEN_CHANNEL(c2)) +
		(BLUE_CHANNEL(c1)-BLUE_CHANNEL(c2)) * (BLUE_CHANNEL(c1)-BLUE_CHANNEL(c2));
	

	return(ds);

}

/* See- https://en.wikipedia.org/wiki/Alpha_compositing */
rgba_t rgba_alphablend(rgba_t rA, rgba_t rB) {

	drgba_t A,B,C;

	rgba_to_drgba(rA,&A);
	rgba_to_drgba(rB,&B);

	C.a = A.a + (B.a*(1-A.a));
	if(C.a > 0.0) {
		C.r = ((A.r*A.a) + (B.r*B.a*(1-A.a)))/C.a;
		C.g = ((A.g*A.a) + (B.g*B.a*(1-A.a)))/C.a;
		C.b = ((A.b*A.a) + (B.b*B.a*(1-A.a)))/C.a;
	} else {
		C.r=C.g=C.b = 0.0;
	}
	return(rgba_from_drgba(&C));

}

/* independendly set the alpha channel of c. */
rgba_t rgba_alpha(rgba_t c,channel_t a) {
	return(c & ((~ALPHA_SHIFT(0xFF)) | ALPHA_CHANNEL(a)));
}

/* map each channel into one of the ansi216 color values found on
 * https://en.wikipedia.org/wiki/ANSI_escape_code. Note that these values
 * aren't spaced out evenly.  You have to look in a bucket to get the right
 * value.  */
rgba_t rgba_ansi216(rgba_t color) {
	rgba_t out = 0;
	channel_t r,g,b,a;
	channel_t bucket[6] = { 0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff };

	rgba_to_channel(color,&r,&g,&b,&a);

	r=bucket[ansi216_index(r)];
	g=bucket[ansi216_index(g)];
	b=bucket[ansi216_index(b)];

	out = rgba_from_channel(r,g,b,a);

	return(out);
}

/* returns a new fg color that tries to get fg and bg within the mincr.  I'm
 * not sure this is guarneteed to actually meet the goal- but it'll get close
 * as it can without changing the background color. */
rgba_t rgba_contrasting_color(rgba_t fg, rgba_t bg, double mincr) {
	
	double cr;
	
	cr = rgba_contrast_ratio(fg,bg);
	if(cr >= mincr) {
		return(fg);
	}

	/* which is better, take fg darker or lighter? */
	if( rgba_contrast_ratio(0xFFFFFFFF,bg) > rgba_contrast_ratio(0x000000FF,bg)) {
		/* make fg lighter. */
		for(;(rgba_opaque(fg)<0xFFFFFFFF) && (rgba_contrast_ratio(fg,bg)<mincr);) {
			fg = rgba_screen(fg,0x0c0c0cFF); /* lighten it up by ~5% */
		}
	} else {
		/* make fg darker. */
		for(;(rgba_opaque(fg)>0x000000FF) && (rgba_contrast_ratio(fg,bg)<mincr);) {
			fg = rgba_multiply(fg,0xf3f3f3FF); /* darken it up by ~5% */
		}
	}
	return(fg);
}

/* return the index from the ansi256 array with a little sanity checking applied. */
rgba_t rgba_from_ansi256(int idx) {
	if(idx > ARRAY_SIZE(ansi256) || idx < 0) {
		return(0xDEADBEEF);
	}
	return(ansi256[idx].color);
}

/* See- https://en.wikipedia.org/wiki/Alpha_compositing */
rgba_t rgba_gammablend(rgba_t A, rgba_t B, double y) {
	rgba_t Ay,By,Cy;

	if(y == 1.0) {
		return(rgba_alphablend(A,B));
	}

	Ay = rgba_gamma(A,1.0/y);
	By = rgba_gamma(B,1.0/y);
	Cy = rgba_alphablend(Ay,By);
	return(rgba_gamma(Cy,y));

}

/* note that gamma is not applied to the alpha channel.*/
rgba_t rgba_gamma(rgba_t c,double y) {

	drgba_t d;

	rgba_to_drgba(c,&d);
	d.r = pow(d.r,y);
	d.g = pow(d.g,y);
	d.b = pow(d.b,y);

	return(rgba_from_drgba(&d));

}

/* convert a color into linear greyscale */
rgba_t rgba_greyscale(rgba_t color) {

	channel_t r,g,b,a;
	int Y; /* luma */

	rgba_to_channel(color,&r,&g,&b,&a);
	Y = (r+b+g)/3.0 * ((double)a/0xff);
	return(rgba_from_channel(Y,Y,Y,a));

}

/* deprecate this! */
rgba_t rgba_hsvblend(rgba_t A, rgba_t B, double amount, int ccw) {

	hsva_t Ah, Bh, Nh;
	double distance;

	if(amount <= 0.0) {
		return(B);
	}
	if(amount >= 1.0) {
		return(A);
	}

	rgba_to_hsva(A,&Ah);
	rgba_to_hsva(B,&Bh);

	if(Ah.h < 0.0) Ah.h +=360.0;
	if(Bh.h < 0.0) Bh.h +=360.0;


	if(!ccw) {
		if(Ah.h == Bh.h) { 
			Bh.h += 360;
		} else {
			if(Bh.h < Ah.h) Bh.h += 360.0;
		}
		distance = Bh.h-Ah.h;
		Nh.h = Ah.h + (1.0-amount)*distance;
		if(Nh.h > 360.0) Nh.h -= 360.0;
	} else {
		if(Ah.h == Bh.h) { 
			Bh.h -= 360;
		} else if(Bh.h > Ah.h) Bh.h -= 360.0;
		distance = Ah.h-Bh.h;
		Nh.h = Ah.h - (1.0-amount)*distance;
		if(Nh.h < 0.0) Nh.h += 360.0;
	}

	Nh.s = ((Ah.s * amount) + (Bh.s * (1.0-amount)));
	Nh.v = ((Ah.v * amount) + (Bh.v * (1.0-amount)));
	Nh.a = ((Ah.a * amount) + (Bh.a * (1.0-amount)));

	return(rgba_from_hsva(&Nh));

}


rgba_t rgba_hsvblend_long(rgba_t A, rgba_t B, double amount) {
	return(rgba_hsvblend_path(A,B,amount,1));
}

rgba_t rgba_hsvblend_short(rgba_t A, rgba_t B, double amount) {
	return(rgba_hsvblend_path(A,B,amount,0));
}

/* perform a blend between two rgba colors by converting to hsva, then
 * 'rotating' the hue from the first color to the second color. Note that there
 * are two ways to do this- the way that 'spins' the least distance and the way
 * that spins the most distance; and a special case when the hues are equal-
 * leave the hue alone, spin the hue in a full circle positive direction, or
 * spin the hue in a full circle in the negative direction.  longway, if true,
 * will prefer spining the hue in the direction that takes the longest path, or
 * if the hues are equal, in the positive direction.  If false, the shortest
 * path is taken, and if the hues are equal, they are not changed.  longway, if
 * true, makes rainbow style blends. */
rgba_t rgba_hsvblend_path(rgba_t A, rgba_t B, double amount, int longway) {

	hsva_t Ah, Bh, Nh;
	double hue,hue1,hue2;
	double d1,d2;

	if(amount <= 0.0) {
		return(B);
	}
	if(amount >= 1.0) {
		return(A);
	}

	rgba_to_hsva(A,&Ah);
	rgba_to_hsva(B,&Bh);

	if(Ah.h < 0.0) Ah.h +=360.0;
	if(Bh.h < 0.0) Bh.h +=360.0;


	/* figure out the first path */
	if(Ah.h == Bh.h) { 
		//hue = Bh.h + 360;
		/* Not bumping the hue up will ensure that if the two hues are the
		 * same, the shortest path is not to change anything. */
		hue = Bh.h;  
	} else if(Bh.h < Ah.h) { 
		hue = Bh.h + 360.0;
	} else {
		hue = Bh.h;
	}
	d1 = hue-Ah.h;
	hue1 = Ah.h + (1.0-amount)*d1;
	if(hue1 > 360.0) hue1 -= 360.0;

	/* and the second path */
	if(Ah.h == Bh.h) { 
		hue = Bh.h - 360;
	} else if(Bh.h > Ah.h) {
		hue = Bh.h - 360.0;
	} else {
		hue = Bh.h;
	}
	d2 = Ah.h-hue;
	hue2 = Ah.h - (1.0-amount)*d2;
	if(hue2 < 0.0) hue2 += 360.0;

	if(abs(d1) > abs(d2)) {
		/* make hue1 the shortest path. */
		hue = hue1;
		hue1 = hue2;
		hue2 = hue;
	}
	if(longway) {
		hue = hue2;
	} else {
		hue = hue1;
	}

	Nh.h = hue;
	Nh.s = ((Ah.s * amount) + (Bh.s * (1.0-amount)));
	Nh.v = ((Ah.v * amount) + (Bh.v * (1.0-amount)));
	Nh.a = ((Ah.a * amount) + (Bh.a * (1.0-amount)));

	return(rgba_from_hsva(&Nh));

}

/* multiply the channels together, including alpha */
rgba_t rgba_invert(rgba_t c) {
	channel_t r,g,b,a;

	rgba_to_channel(c,&r,&g,&b,&a);
	return(rgba_from_channel((0xff-r),(0xff-g),(0xff-b),(0xff-a)));

}

/* multiply the channels together, including alpha */
rgba_t rgba_multiply(rgba_t a, rgba_t b) {
	channel_t ar,ag,ab,aa;
	channel_t br,bg,bb,ba;

	rgba_to_channel(a,&ar,&ag,&ab,&aa);
	rgba_to_channel(b,&br,&bg,&bb,&ba);
	return(rgba_from_channel(ar*br/0xff,ag*bg/0xff,ab*bb/0xff,aa*ba/0xff));

}

/* set the alpha channel to full opaque. */
rgba_t rgba_opaque(rgba_t c) {
	return( c | ALPHA_SHIFT(0xff));
}

/* map each channel into one of N distinct values. */
rgba_t rgba_posterize(rgba_t color, double N) {
	rgba_t out = 0;
	double r,g,b,a;

	rgba_to_double(color,&r,&g,&b,&a);
	r=posterize_double(r,N);
	g=posterize_double(g,N);
	b=posterize_double(b,N);
	a=posterize_double(a,N);
	out = rgba_from_double(r,g,b,a);

	return(out);
}

/* multiply the channels together, including alpha */
rgba_t rgba_screen(rgba_t a, rgba_t b) {
	a = rgba_invert(a);
	b = rgba_invert(b);
	return(rgba_invert(rgba_multiply(a,b)));
}

/* map a value c from range 0.0-1.0 into one of N distinct values. */
double posterize_double(double c,double N) {
	if(N <= 0.0) return(c);
	return(floor(c*N)/N);
}

/* return the euclidian squared distance of two rgba colors. Alpha ignored.*/
double drgba_distance2(drgba_t *c1, drgba_t *c2) {
	double ds;

	ds= ((c1->r)-(c2->r)) * ((c1->r)-(c2->r)) +
		((c1->g)-(c2->g)) * ((c1->g)-(c2->g)) +
		((c1->b)-(c2->b)) * ((c1->b)-(c2->b));

	return(ds);

}

/* ANSI Terminal code functions. */

/* calculate and send the 6x6x6 color cube index directly from the rgba color,
 * no closest matching algorithm involved. */
int sgr_ansi216(char *buf, size_t size, rgba_t color, rgba_context *ctx) {

	int len = 0;
	channel_t ri,gi,bi;
	rgba_t sixbit;
	int index;

	ctx = resolve_ctx(ctx);

	/* sixbit is the color chosen from the ansi ansi216. */
	sixbit = rgba_ansi216(color);
	rgba_to_ansi216idx(sixbit,&ri,&gi,&bi,NULL);
	index = 16 + (36 * ri) + (6 * gi) + (bi);

	len = g_snprintf(buf,size,"%s[%d;5;%dm",
		ESC(ctx),
		(ctx->fgbg_mode==ANSI_BG)?48:38,
		index
	);

	return(len);

}

/* send the set graphic rendition sequence to enable or disable a particular
 * individual character attribute. This checks ctx->current_style.attr and will not
 * reassert an attribute if it is alreay enabled. */
int sgr_ansiAttr(char *dst, size_t dlen, ansi_ca attr, gboolean state, rgba_context *ctx) {

	char *d=dst;
	char *eod=dst+dlen;

	ctx = resolve_ctx(ctx);

	if(ctx->color_mode == ANSI_MONO) {
		if(dlen>=1) *d='\0';
		return(0);
	}

	/* is attr known? */
	if((attr<0) || (attr > ANSI_CA_MAX)) {
		if(dlen>=1) *d='\0';
		return(0);
	}

	/* Do not reassert. */
	if(ctx->current_style.attr[attr] == state) {
		if(dlen>=1) *d='\0';
		return(0);
	}

	d+=g_snprintf(d,eod-d,"%s[%sm",
		ESC(ctx),
		ansi_attributes[attr].Ps[(state==TRUE)?1:0]
	);
	ctx->current_style.attr[attr] = state;

	if( (attr != ANSI_CA_NORMAL) &&
		(state == TRUE)
	) {
		ctx->current_style.attr[ANSI_CA_NORMAL] = FALSE;
	}

	return(d-dst);

}

/* Insert the closest match ANSI Set Graphic Rendition sequence for a given
 * rgba color and color mode into buf.  When currentcolor is non-null, the sgr
 * sequence is supressed when the closest match is the same as currentcolor.
 * Otherwise, currentcolor is updated to the closest match, and the sgr
 * sequence is copied into buf.  This is useful for reducing redundant color
 * change sgr sequences, which can add up very quickly in range blended texts.
 * */
int sgr_ansiColor(char *buf, size_t size, rgba_t color, rgba_context *ctx) {

	int from,to;
	colorchip_t *match;
	int bold,rev;
	char *s,*eos;
	rgba_t contrasting;

	ctx = resolve_ctx(ctx);
	s = buf;
	eos = buf+size;

	if(ctx->fgbg_mode == ANSI_FG) {
		ctx->unmodified_fg = color;
	}

	if(ALPHA_CHANNEL(color) != 0xff) {
		switch(ctx->fgbg_mode) {
			case ANSI_FG:
				/* fg blends with current bg. */
				color = rgba_opaque(rgba_gammablend(color,ctx->current_style.bg,ctx->gamma));
				break;
			case ANSI_BG:	
				/* bg blends with default bg. */
				color = rgba_opaque(rgba_gammablend(color,ctx->default_style.bg,ctx->gamma));
				break;
			default:
				break;
		}
	}

	if(ctx->contrast_ratio != 1.0) {
		switch(ctx->fgbg_mode) {
			case ANSI_FG:
				/* modifiy the requested color to contrast with current background. */
				color = rgba_contrasting_color(color,ctx->current_style.bg,ctx->contrast_ratio);
				break;
			case ANSI_BG: 
				/* if the current fg color isn't the previous unmodified one, then send a new fg code. */
				contrasting = rgba_contrasting_color(ctx->unmodified_fg,color,ctx->contrast_ratio);

				if(contrasting != ctx->current_style.fg) {
					/* want to maintain the existing unmodified_fg and current_bg! */
					rgba_t unmodified_fg = ctx->unmodified_fg;
					rgba_t current_bg = ctx->current_style.bg;
					ctx->fgbg_mode = ANSI_FG;
					ctx->current_style.bg = color;
					s += sgr_ansiColor(s,eos-s,contrasting,ctx);
					ctx->current_style.bg=current_bg;
					ctx->unmodified_fg = unmodified_fg;
					ctx->fgbg_mode = ANSI_BG;
				}
				break;
			default:
				break;
		} 
	}

	switch (ctx->color_mode) {
		default:
		case ANSI_MONO:
			*s='\0';
			return(s-buf);

		case ANSI_TRUECOLOR:
			/* supplied color is already the true target color, so there is no
			 * need to do any other palette matching.  let sgr_ansiTruecolor()
			 * do the work. */
			if(is_current_color(color,ctx)) {
				*s='\0';
			} else {
				set_current_color(color,ctx);
				s += sgr_ansiTruecolor(s,eos-s,color,ctx);
			}
			return(s-buf);

		/* the following cases are all coming from the indexed ANSI-256 palette. */
		case ANSI_216:
			/* Might be faster to get the six cube color algorithmically with
			 * sgr_ansi216() instead.... but I'm not doing it that way.  Best
			 * matching it simplifies the ANSI_240 and ANSI_256 cases.*/
			/* return(sgr_ansi216(buf,size,color,fgbg)); */
			from = 16;
			to = 232;
			break;
		case ANSI_STD:
		case ANSI_16:
			from = 0;
			to = 15;
			break;
		case ANSI_GREY:
			from = 232;
			to = 255;
			break;
		case ANSI_240:
			from = 16;
			to = 255;
			break;
		case ANSI_256:
			from = 0;
			to = 255;
			break;
	}

	if(!(match=rgba_closest_match(color,ansi256,from,to))) {
		/* the match failed for some reason? */
		*s='\0';
		return(s-buf);
	}

	/* handle supression */
	if(is_current_color(match->color,ctx)) {
		*s='\0';
		return(s-buf);
	} else {
		set_current_color(match->color,ctx);
	}

	switch(ctx->color_mode) {
		case ANSI_STD: 
			/* a little more remapping is needed for STD colors. */
			bold=0;
			rev=0;
			/* The 8 colors in the STD palette begin at SGR 30 */
			int sgr= match->index + 30;

			if(match->index >= 8) {
				/* The 8 additional colors in the STD palette are rendered by
				 * sending SGR bold, then one of of the first 8 colors. */
				bold = 1;
				sgr = match->index + 30 - 8;
			}
			/* In STD mode, you can't really change the BG color independednly of
			 * the FG, but we can fake it by turning on SGR reverse mode. */
			if(ctx->fgbg_mode == ANSI_BG) {
				rev=1;
			}

			/* now send the combinded SGR parameters. */
			s+= g_snprintf(s,eos-s,"%s[%s%s%dm", 
				ESC(ctx),
				(bold)?"1;":"22;", /* set bright mode? */
				(rev)?"7;":"27;", /* set reverse mode? */
				sgr
			);
			break;

		default:
		case ANSI_GREY: 
		case ANSI_16: 
		case ANSI_216: 
		case ANSI_240: 
		case ANSI_256: 
			/* the other color modes are all sent as mode 5 (256 color palette) indexes. */
			s+=g_snprintf(s,eos-s,"%s[%d;5;%dm",
				ESC(ctx),
				(ctx->fgbg_mode==ANSI_BG)?48:38,
				match->index
			);
		
			break;
	}
	return(s-buf);

}

/* convenience function- ansi reset, all attributes off. */
int sgr_ansiNormal(char *buf, size_t size, rgba_context *ctx) {
	ctx = resolve_ctx(ctx);
	for(int i=0;i<ANSI_CA_MAX;i++) {
		ctx->current_style.attr[i] = FALSE;
	}
	ctx->current_style.attr[ANSI_CA_NORMAL] = TRUE;
	ctx->unmodified_fg = ctx->default_style.fg;
	ctx->current_style.fg = ctx->default_style.fg;
	ctx->current_style.bg = ctx->default_style.bg;
	return(g_snprintf(buf,size,"%s[0m",ESC(ctx)));
}

/* send the correct sequence to apply the style. */
int sgr_ansiStyle(char *dst, size_t dlen, style_t *style, rgba_context *ctx) {

	char *d=dst;
	char *eod=dst+dlen;
	int needsupdate=0;
	ansi_fgbg tmpfgbg;

	ctx = resolve_ctx(ctx);
	tmpfgbg = ctx->fgbg_mode;

	/* loop over the attributes first, since the 'normal' attribute can reset
	 * the color directives, and let sgr_ansiAttr do the work of figuring out
	 * if the option really needs to be sent. */
	if(style->opt[STYLE_ATTR]) {
		for(int i=0;i<ANSI_CA_MAX;i++) {
			d+=sgr_ansiAttr(d,eod-d,i,style->attr[i],ctx);
		}
	}

	/* send the bg first, so that fg can hit the right contrast ratio. */
	if(style->opt[STYLE_BG]) {
		ctx->fgbg_mode = ANSI_BG;
		d+=sgr_ansiColor(d,eod-d,style->bg,ctx);
	}

	/* send the fg. */
	if(style->opt[STYLE_FG]) {
		ctx->fgbg_mode = ANSI_FG;
		d+=sgr_ansiColor(d,eod-d,style->fg,ctx);
	}

	ctx->fgbg_mode = tmpfgbg;
	return(d-dst);

}

/* print an ansi terminal control sequence into buf that uses the color to set
 * either the forground or the background.  Note that the alpha channel is
 * ignored here, since ansi doesn't have a concept of that.  */
int sgr_ansiTruecolor(char *buf, size_t size, rgba_t color, rgba_context *ctx) {

	int ret = 0;
	channel_t r,g,b;
	int modecode;

	switch(ctx->fgbg_mode) {
		default:
		case ANSI_FG:
			modecode = 38;
			break;
		case ANSI_BG:
			modecode = 48;
			break;
	}

	rgba_to_channel(color,&r,&g,&b,NULL);
	ret = g_snprintf(buf,size,"%s[%2d;2;%d;%d;%dm",
		ESC(ctx),
		modecode,
		r,g,b
	);
	
	return(ret);

}

/* convenience function- ansi reset, re-issue context defaults. */
int sgr_ctxDefaults(char *buf, size_t size, rgba_context *ctx) {
	
	char *s,*eos;
	s = buf;
	eos = buf+size;

	ctx = resolve_ctx(ctx);
	s+=sgr_ansiNormal(s,eos-s,ctx);

	/* send default colors. */
	ctx->fgbg_mode = ANSI_BG;
	set_current_color(~ctx->default_style.bg,ctx);
	s+=sgr_ansiColor(s,eos-s,ctx->default_style.bg,ctx);
	ctx->fgbg_mode = ANSI_FG;
	set_current_color(~ctx->default_style.fg,ctx);
	s+=sgr_ansiColor(s,eos-s,ctx->default_style.fg,ctx);
	return(s-buf);
}

/* return a printable version of ESC if context is in test mode. */
const char *ESC(rgba_context *ctx) {
	if(ctx->debug) {
		return("␛");
	}
	return("\x1b");
}

/* Given a single R,G,B color channel in the range 0-0xff, return the 6x6x6
 * color cube index in the range 0-5.
 * https://github.com/tmux/tmux/blob/master/colour.c */
int ansi216_index(channel_t ch) {
	if (ch < 0x30)
		return (0);
	if (ch < 0x72)
		return (1);
	return ((ch - 0x23) / 0x28);
}

/* tests terminal is already sending this color. */
int is_current_color(rgba_t color,rgba_context *ctx) {

	switch(ctx->fgbg_mode) {
		default:
		case ANSI_FG:
			if(color == ctx->current_style.fg) return(1);
			break;
		case ANSI_BG:
			if(color == ctx->current_style.bg) return(1);
			break;
	}
	return(0);

}

/* Sets current color in the context.  Returns the value previously in use. */
rgba_t set_current_color(rgba_t color,rgba_context *ctx) {
	rgba_t oldcolor;
	switch(ctx->fgbg_mode) {
		case ANSI_FG:
			oldcolor = ctx->current_style.fg;
			ctx->current_style.fg = color;
			return(oldcolor);
		case ANSI_BG:
			oldcolor = ctx->current_style.bg;
			ctx->current_style.bg = color;
			return(oldcolor);
		default:
			break;
	}
	return(0xDEADBEEF);

}

/* prints out the six cube colors in the palette entry format.  (use this to
 * print out the mid range values of the ansi256 palette for inclusion in the
 * source) */
void printansi216(void) {
	int r,g,b;
	int index;
	rgba_t color;
	channel_t bucket[6] = { 0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff };

	for(r=0;r<=5;r++) {
		for(g=0;g<=5;g++) {
			for(b=0;b<=5;b++) {
				index = 16 + (36 * r) + (6 * g) + (b);
				color = rgba_from_channel(bucket[r],bucket[g],bucket[b],0xff);
				fprintf(stdout,"{ \"six%d%d%d\", 0x%08X, %d}, %c",
					r,g,b,
					color,
					index,
					(b&1)?'\n':' '
				);
			}
		}
	}
}

/* palette related functions */

/* look up a named color in the palette pal, or the system palette if pal is
 * NULL.   if found, set color and return 1, else return 0. */
int lookup_color_name(char *name, rgba_context *ctx, rgba_t *color) {

	colorchip_t *entry = NULL;

	if(!name) return (0);

	ctx = resolve_ctx(ctx);

	if( (ctx->palette) && 
		(entry = g_hash_table_lookup(ctx->palette,name)) 
	) {
		if(color) *color = entry->color;
		return(1);
	} 
	if( ctx->system_palette_fallback ) {
		if( (default_ctx->palette) && 
			(entry = g_hash_table_lookup(default_ctx->palette,name)) 
		) {
			if(color) *color = entry->color;
			return(1);
		}
	}
	if(color) *color = 0xDEADBEEF;
	return(0);
}

/* load a palette (name color) from a file, return number of unique entries.
 * Additional calls will merge files into the palette name hash. */
int load_colorchip_hash(char *filename, rgba_context *ctx) {

	int count = 0;
	char name[NAMED_COLOR_LEN];
	char colortext[NAMED_COLOR_LEN];
	char line[MAX_STRING_LENGTH];
	rgba_t color;
	colorchip_t *entry;
	FILE *in;

	if(!filename) return(-1);
	ctx = resolve_ctx(ctx);

	if(ctx->palette == NULL) {
		ctx->palette = new_palette();
	}

	if(!(in=fopen(filename,"r"))) {
		return(-1);
	}

	while( fgets(line,sizeof(line),in) ) {
		if((sscanf(line,"%64s %64s",name,colortext) == 2) ) {
			if( *name == '#' ) continue;
			if( (rgba_sscanf(colortext,&color,ctx) )) {
				entry = new_colorchip();
				entry->name = g_strdup(name);
				entry->color = color;
				/* No plan to use the index for anything, but I'm setting it to
				 * ansi red in the 256k palette so that is is visible in case it
				 * gets used by mistake. */
				entry->index = 1;  
				if ( (g_hash_table_insert(ctx->palette,g_strdup(name),entry) )) {
					count++;
				}
			}
		}
	}

	fclose(in);
	return(count);
}

/* parse buf into a color value.  Returns 1 if found, 0 if not parsable. */ 
int rgba_sscanf(char *buf,rgba_t *c, rgba_context *ctx) {
	int r,g,b,a;
	char name[NAMED_COLOR_LEN];
	colorchip_t *lc;
	if(!buf) return(0);

	/* #rrggbbaa */
	if(sscanf(buf,"#%2x%2x%2x%2x",&r,&g,&b,&a) == 4) {
		if(c) *c = rgba_from_channel(r,g,b,a);
		return(1);
	}

	/* #rrggbb */
	if(sscanf(buf,"#%2x%2x%2x",&r,&g,&b) == 3) {
		if(c) *c = rgba_from_channel(r,g,b,0xff);
		return(1);
	}

	/* Shorthand style, #rgba */
	if(sscanf(buf,"#%1x%1x%1x%1x",&r,&g,&b,&a) == 4) {
		if(c) *c = rgba_from_channel(r*16+r,g*16+g,b*16+b,a*16+a);
		return(1);
	}

	/* Shorthand style, #rgb */
	if(sscanf(buf,"#%1x%1x%1x",&r,&g,&b) == 3) {
		if(c) *c = rgba_from_channel(r*16+r,g*16+g,b*16+b,0xff);
		return(1);
	}

	/* DefaultFG is reserved color name.*/
	if(!strcasecmp(buf,"DefaultFG")) {
		if(c) *c = ctx->default_style.fg;
		return(1);
	}

	/* DefaultBG is reserved color name.*/
	if(!strcasecmp(buf,"DefaultBG")) {
		if(c) *c = ctx->default_style.bg;
		return(1);
	}

	/* look for it by name in the supplied table. */
	ctx = resolve_ctx(ctx);
	if(sscanf(buf,"%64s",name) == 1) {
		if( (ctx->palette) && 
			(lc = (colorchip_t *)g_hash_table_lookup(ctx->palette,name)) 
		) {
			if(c) *c = lc->color;
			return(1);
		}
		if(ctx->system_palette_fallback) { 
			if( (default_ctx->palette) &&
				(lc = (colorchip_t *)g_hash_table_lookup(default_ctx->palette,name)) 
			) {
				if(c) *c = lc->color;
				return(1);
			}
		}
	}
	return(0);

}

/* parse buf into a color value.  Returns 1 if found, 0 if not parsable. */ 
int colorchip_sscanf(char *buf,colorchip_t *c, rgba_context *ctx) {
	int r,g,b,a;
	char name[NAMED_COLOR_LEN];
	colorchip_t *lc;
	if(!buf) return(0);
	if(!c) return(0);

	if(!rgba_sscanf(buf,&(c->color),ctx)) {
		return(0);
	}

	if(c->name) {
		g_free(c->name);
	}
	c->name = g_strdup(buf);
	c->index = 0x0;  /* not right. */
	return(1);

}

colorchip_t *new_colorchip(void) {

	colorchip_t *new;
	new = (colorchip_t *)malloc(sizeof(colorchip_t));
	new->name = NULL;
	new->definition = NULL;
	new->color = 0x00000000;
	new->index = 0;
	return(new);
}

colorchip_t *new_colorchip_parse(char *name,rgba_context *ctx) {

	colorchip_t *new;
	ctx = resolve_ctx(ctx);
	new = new_colorchip();

	colorchip_sscanf(name,new,ctx);
	return(new);

}

/* simple linear search through the palette. From and to determine the starting
 * and ending array indexes of the palette, so that subset ranges of colors can
 * be scanned. This is faster than rgba_closest_match_double, with no
 * signifcant loss of precision. */
colorchip_t *rgba_closest_match(rgba_t color, colorchip_t *pal, int from, int to) {
	
	int dbest = 0xffffff;  /* larger than the largest rgba_distance2 */
	int dist;
	colorchip_t *best = NULL;
	colorchip_t *test;

	test = pal+from;
	for(int i=0;i<(to-from);i++) {
		if(color == test->color) {
			/* exact match */
			return(test);
		}
		dist = rgba_distance2(color,test->color);
		if(dist < dbest) {
			dbest = dist;
			best = test;
		}
		test++;
	}
	return(best);
}

/* simple linear search through the palette. From and to determine the starting
 * and ending array indexes of the palette, so that subset ranges of colors can
 * be scanned. */
colorchip_t *rgba_closest_match_double(rgba_t color, colorchip_t *pal, int from, int to) {
	
	double dbest = 5;  /* larger than the largest rgba_distance2 */
	double dist;
	colorchip_t *best = NULL;
	colorchip_t *test;
	drgba_t dcolor,dtest;

	rgba_to_drgba(color,&dcolor);

	test = pal+from;
	for(int i=0;i<(to-from);i++) {
		if(color == test->color) {
			/* exact match */
			return(test);
		}
		rgba_to_drgba(test->color,&dtest);
		dist = drgba_distance2(&dcolor,&dtest);
		if(dist < dbest) {
			dbest = dist;
			best = test;
		}
		test++;
	}
	return(best);
}


/* make a new palette for name to color lookup */
colorchip_hash_t *new_palette(void) {
	colorchip_hash_t *new = NULL;
	/* fully define the hash table functions at creation, so destruction can
	 * be a simple g_hash_table_destroy() later. */
	new = g_hash_table_new_full(
		g_str_hash,     /* standard string hash function. */
		g_str_equal,    /* standard string compare function. */
		g_free,         /* standard string free function for the key. */
		(GDestroyNotify) free_colorchip  /* non-standard free function for the datatype. */
	);
	return(new);
}

/* look up a color by string spec or by name in the ctx palette. */
rgba_t rgba_parse(char *name,rgba_context *ctx) {
	rgba_t color;

	if(rgba_sscanf(name,&color,ctx)) {
		return(color);
	} else {
		return(0xDEADBEEF);
	}
}

void free_colorchip(colorchip_t *f) {
	if(!f) return;
	if(f->name) g_free(f->name);
	if(f->definition) g_free(f->definition);
	free(f);
}

/* free a color name lookup table.*/
void free_colorchip_hash(colorchip_hash_t *f) {
	/* see? I told you it would be simple. */
	if(f) g_hash_table_destroy(f);
}

/* style related functions */

/* load styles (name style) from a file, return number of unique entries.
 * Additional calls will merge files into the palette name hash. */
int load_style_hash(char *filename, rgba_context *ctx) {

	int count = 0;
	char name[NAMED_STYLE_LEN];
	char styletext[NAMED_STYLE_LEN];
	char line[MAX_STRING_LENGTH];
	style_t *style;
	style_t *entry;
	FILE *in;

	if(!filename) return(-1);
	ctx = resolve_ctx(ctx);

	if(ctx->styles == NULL) {
		ctx->styles = new_style_hash();
	}

	if(!(in=fopen(filename,"r"))) {
		return(-1);
	}

	while( fgets(line,sizeof(line),in) ) {
		if((sscanf(line,"%64s %512s",name,styletext) == 2) ) {
			if( *name == '#' ) continue;
			/* create a new entry */
			entry = new_style();
			if( (style_sscanf(styletext,&entry,ctx) )) {
				/* if sscanf was able to parse the styletext, annotate and save the entry. */
				if(entry->name) free(entry->name);
				entry->name = g_strdup(name);
				if(entry->definition) free(entry->definition);
				entry->definition = g_strdup(styletext);
				if ( (g_hash_table_insert(ctx->styles,g_strdup(name),entry) )) {
					/* note- if a same named entry already existed, that one
					 * will be free'd by the hash table insert, and the new one
					 * gets added */
					count++;
				}
			} else {
				/* sscanf wasn't able to parse the entry. Toss it back. */
				free_style(entry);
			}
		}
	}

	fclose(in);
	return(count);
}

/* scan for a style string.  format is...
 * <fgcolor>[/<bgcolor>[/<attrname>[,<attrname>]]]
 */
int style_sscanf(char *buf, style_t **styleloc, rgba_context *ctx) {

	char field[MAX_STRING_LENGTH];
	rgba_t color;
	char *b, *f;
	int len=0;
	style_t *style;

	if(!buf || *buf=='\0') return(0);
	if(!styleloc) return(0);

	ctx = resolve_ctx(ctx);

	/* if the buffer contains no /'s, this is either annother style definition,
	 * or a style with a single fg color.  A pre-defined style wins out. */
	if(!(strchr(buf,'/'))) {
		if((style=lookup_style_name(buf,ctx))) {
			style_cpy(*styleloc,style);
			return(1);
		}
	}

	/* start of with a copy of the context's default style. We'll copy into the
	 * right location further down.*/
	style = style_dup(&(ctx->default_style));

	/* clear out all the style options, they'll add back in as they are
	 * detected. */
	for(int i=0;i<STYLE_MAX;i++) {
		style->opt[i] = FALSE;
	}
	
	b=buf;
	f=field;
	while(*b && *b!='/') {
		*f++=*b++;
	}
	*f++='\0';
	if( rgba_sscanf(field,&color,ctx) ) {
		style->fg = color;
		style->opt[STYLE_FG] = TRUE;
	} else if ( *b == '\0' ) {
		/* it wasn't a valid color either. */
		free_style(style);
		return(0);
	}
	if(*b == '/') {
		b++;
		f=field;
		while(*b && *b!='/') {
			*f++=*b++;
		}
		*f++='\0';
		if( rgba_sscanf(field,&color,ctx) ) {
			style->bg = color;
			style->opt[STYLE_BG] = TRUE;
		}
	}
	if(*b == '/') {
		int atcnt = 0;
		b++;
		while(*b) {
			f=field;
			while(*b && *b!=',') {
				*f++=*b++;
			}
			*f++='\0';
			len = strlen(field);
			for(int i=0;i<ANSI_CA_MAX;i++) {
				if(!strncasecmp(ansi_attributes[i].name,field,len)) {
					style->attr[i] = TRUE;
					style->opt[STYLE_ATTR] = TRUE;
					if(i!=ANSI_CA_NORMAL) atcnt++;
					break;
				}
			}
			if(*b == ',') b++;
		}
		if(atcnt>=1) {
			style->attr[ANSI_CA_NORMAL] = FALSE;
		}
	}

	/* copy the style we built into the location provided by the caller.*/
	style_cpy(*styleloc,style);
	/* ...and don't forget to clean up our mess.*/
	free_style(style);
	return(1);
}

/* make a new palette for name to color lookup */
style_hash_t *new_style_hash(void) {
	style_hash_t *new = NULL;
	/* fully define the hash table functions at creation, so destruction can
	 * be a simple g_hash_table_destroy() later. */
	new = g_hash_table_new_full(
		g_str_hash,     /* standard string hash function. */
		g_str_equal,    /* standard string compare function. */
		g_free,         /* standard string free function for the key. */
		(GDestroyNotify) free_style /* non-standard free function for the datatype. */
	);
	return(new);
}

/* look up a named style in the ctx, or the system palette if ctx or
 * ctx->styles is NULL.   if found, set color and return 1, else return 0. */
style_t *lookup_style_name(char *name, rgba_context *ctx) {

	style_t *entry = NULL;

	if(!name) return (0);

	ctx = resolve_ctx(ctx);

	if( (ctx->styles) && 
		(entry = g_hash_table_lookup(ctx->styles,name)) 
	) {
		return(entry);
	} 
	if( ctx->system_styles_fallback ) {
		if( (default_ctx->styles) && 
			(entry = g_hash_table_lookup(default_ctx->styles,name)) 
		) {
			return(entry);
		}
	}
	return(NULL);
}

style_t *new_style(void) {

	style_t *new;
	new = (style_t *)g_malloc(sizeof(style_t));
	new->name = NULL;
	new->definition = NULL;
	new->fg = 0xDEADBEEF;
	new->bg = 0x0;
	for(int i=0;i<ANSI_CA_MAX;new->attr[i++]=0);
	new->attr[ANSI_CA_NORMAL] = TRUE;
	for(int i=0;i<STYLE_MAX;new->opt[i++]=TRUE);

	return(new);
}

/* create a new style from the style string definition.*/
style_t *new_style_parse(char *name,rgba_context *ctx) {
	
	style_t *new;
	ctx = resolve_ctx(ctx);
	new = new_style();
	
	style_sscanf(name,&new,ctx);
	return(new);
}

/* deep-copy s into d, freeing and duping internal items as needed. */
style_t *style_cpy(style_t *d, style_t *s) {

	if(!s) return(NULL);
	if(!d) return(NULL);

	if(d->name) free(d->name);
	d->name = g_strdup(s->name);
	if(d->definition) free(d->definition);
	d->definition = g_strdup(s->definition);
	d->fg = s->fg;
	d->bg = s->bg;
	for(int i=0;i<ANSI_CA_MAX;i++) d->attr[i] = s->attr[i];
	for(int i=0;i<STYLE_MAX;i++) d->opt[i] = s->opt[i];

	return(d);
}

/* return a new private copy of an existing style type. */
style_t *style_dup(style_t *s) {
	style_t *n;
	if(!s) return(NULL);

	n=new_style();
	return(style_cpy(n,s));
}

void free_style(style_t *f) {
	if(!f) return;
	if(f->name) g_free(f->name);
	if(f->definition) g_free(f->definition);
	free(f);
}

/* free a color name lookup table.*/
void free_style_hash(style_hash_t *f) {
	/* see? I told you it would be simple. */
	if(f) g_hash_table_destroy(f);
}

/* fancy style blending and formating functions */

/* Apply a GList of styles across the src text, using blendfn.  Blenfn may be
 * NULL to indicate 'use the default'. */
int style_blend_text(char *dst,size_t dlen,char *src, GList *styles, style_blendfn_t *blendfn, rgba_context *ctx) {

	ansi_fgbg restorefgbg;
	char *s,*d,*eod;
	char *vsrc = NULL;
	const gchar* end;
	double amount;
	gchar utf8seq[7];
	int inlen, x;
	int utf8len;
	rgba_t fg,bg;
	style_t *blended;

	if(!dst) return(0);
	if(!src) return(0);
	if(dlen==0) return(0);

	s = src;
	d = dst;
	eod = dst+dlen;
	ctx = resolve_ctx(ctx);

	restorefgbg = ctx->fgbg_mode;

	/* g_utf8 commands really hate working on non-validated strings, which can
	 * occur when buffers are mistakenly clipped across utf-8 codepoints.  This
	 * prevents subsequent calls from crashing if a clipped point is
	 * encountered. */
	if(!g_utf8_validate (src,-1,&end) ){
		s=vsrc=g_utf8_make_valid(src,end-src);
	}
	inlen = g_utf8_strlen(s,-1);

	for(x=0;*s && x<inlen;x++) {

		amount = (double)x/(inlen-1);

		blended = style_blendfn_list(amount,styles,blendfn,ctx);

		d+=sgr_ansiStyle(d,eod-d,blended,ctx);

		/* print the utf8-char */
		utf8len = g_unichar_to_utf8( g_utf8_get_char(s), utf8seq);
		utf8seq[utf8len]='\0';
		s += utf8len;
		d += g_snprintf(d,eod-d,"%s",utf8seq);
		
		free_style(blended);
	}

	if(vsrc) g_free(vsrc);
	ctx->fgbg_mode = restorefgbg;

	return(d-dst);
}

/* A good basic setup for more complex blending functions.  Basically returns a
 * new style copied from a or b where (amount<=0.5)?a:b. */
style_t *style_blendfn_generic( style_t *a, style_t *b, double amount, rgba_context *ctx) {

	style_t *new;
	ctx = resolve_ctx(ctx);

	/* general behavior for 0.0 and 1.0 */
	if(amount <=0.0) {
		return(style_dup(a));
	} else if (amount >=1.0) {
		return(style_dup(b));
	}
	/* start off with the closer of the two styles */
	if(amount <= 0.5) {
		new = style_dup(a);
	} else {
		new = style_dup(b);
	}
	/* assert the attribute if either a or b has it asserted.*/
	for(int i=0;i<STYLE_MAX;i++) {
		new->opt[i] = (a->opt[i] || b->opt[i]);
	}
	return(new);
}

/* return a new style that is a blend between two other styles from the list.
 * amount is between 0-1.  0 is the fisrt style in the list 1 is the last style
 * in the list, anything in between is a blend.  Caller must remember to free
 * the new style when done with it! */
style_t *style_blendfn_list(double amount,GList *slist, style_blendfn_t *blendfn,rgba_context *ctx) {

	style_t *new;
	style_t *a,*b;
	int aidx,bidx;
	double subamount;

	ctx = resolve_ctx(ctx);

	if(!slist) {
		return(style_dup(&default_style_entry));
	}

	if(blendfn == NULL) {
		blendfn = style_blendfn_gamma;
	}

	int len = g_list_length(slist);
	if(len == 1 || amount < 0.0) {
		return(style_dup(g_list_nth_data(slist,0)));
	}

	double actamt = amount*(len-1);
	aidx = floor(actamt);
	bidx = MIN(aidx+1,len-1);
	subamount = (actamt-aidx);

	a = g_list_nth_data(slist,aidx);
	b = g_list_nth_data(slist,bidx);

	/* the two styles are chosen from the list.. use the requested blend function to grab a new style. */
	new = (blendfn)(a,b,subamount,ctx);

	return(new);
}

/* gamma adjusted linear rgba blender. */
style_t *style_blendfn_gamma( style_t *a, style_t *b, double amount, rgba_context *ctx) {
	style_t *new;

	ctx = resolve_ctx(ctx);
	new = style_blendfn_generic(a,b,amount,ctx);

	new->bg = rgba_gammablend(rgba_alpha(b->bg,(ALPHA_CHANNEL(b->bg)*amount)),a->bg,ctx->gamma);
	new->fg = rgba_gammablend(rgba_alpha(b->fg,(ALPHA_CHANNEL(b->fg)*amount)),a->fg,ctx->gamma);

	return(new);
}

/* short-path hsva style blender. (Red to Yellow goes through orange) */
style_t *style_blendfn_hsva( style_t *a, style_t *b, double amount, rgba_context *ctx) {
	style_t *new;

	ctx = resolve_ctx(ctx);
	new = style_blendfn_generic(a,b,amount,ctx);

	new->bg = rgba_hsvblend_short(b->bg,a->bg,amount);
	new->fg = rgba_hsvblend_short(b->fg,a->fg,amount);

	return(new);
}

/* long-path hsva style blender. (Red to Yellow goes through violet, indigo, blue, green) */
style_t *style_blendfn_rainbow( style_t *a, style_t *b, double amount, rgba_context *ctx) {
	style_t *new;

	ctx = resolve_ctx(ctx);
	new = style_blendfn_generic(a,b,amount,ctx);

	new->bg = rgba_hsvblend_long(b->bg,a->bg,amount);
	new->fg = rgba_hsvblend_long(b->fg,a->fg,amount);

	return(new);
}



/* local other useful code. */








