/* locolor.h - ANSI Terminal Color Handler */
/* Created: Tue 07 Mar 2023 10:34:32 PM EST malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: locolor.h,v 1.12 2023/05/16 13:42:19 malakai Exp $ */

/* Copyright © 2023 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of Locolor - Last Outpost terminal color library
 *
 * Locolor is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * Locolor is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef LOCOLOR_H
#define LOCOLOR_H

/* ---- global #defines ----*/

/* ---- structs and typedefs ---- */

/* a 32 bit unsigned int packed RGBA */
typedef uint32_t rgba_t;  

/* an 8 bit unsigned int for a single color channel */
typedef uint8_t channel_t;

/* an 16 bit unsigned int for a attribute bitvector */
typedef uint16_t attribute_t;

/* a struct of double presision HSVA channels. */
typedef struct {
	double h,s,v,a;
} hsva_t;

/* a struct of double presision RGBA channels. */
typedef struct {
	double r,g,b,a;
} drgba_t;

/* colorchip_hash_t is just a GHashTable. */
typedef GHashTable colorchip_hash_t;

/* style_hash_t is just a GHashTable. */
typedef GHashTable style_hash_t;

/* style_list_t is just a GList. */
typedef GList	style_list_t;

/* Typedef for terminal color selection */
typedef enum {
	ANSI_MONO, /* Don't use any color. */
	ANSI_STD,  /* Original 8 + 8 bold/bright sent as SGR list */
	ANSI_16,   /* The first 16 colors from the mode 5 8-Bit color palette */
	ANSI_GREY, /* The 24 shades of grey from the mode 5 8-Bit color palette */
	ANSI_216,  /* 216 colors from the 6x6x6 mode 5 8-Bit color palette */
	ANSI_240,  /* ANSI_216 + ANSI_GREY using closest match */
	ANSI_256,  /* ANSI_240 + ANSI_16 using closest match */
	ANSI_TRUECOLOR,  /* 24 Bit color sent using SGR mode 2 */
	ANSI_COLOR_MODE_MAX  /* a useful marker. */
} ansi_color_mode;

/* Flag for indicating foreground or background color modification. */
typedef enum {
	ANSI_FG,		/* set only the foreground color */
	ANSI_BG,		/* set only the background color */
	ANSI_FGBG_MAX,	/* a useful marker. */
} ansi_fgbg;

/* text attributes enum */
typedef enum {
	ANSI_CA_NORMAL,
	ANSI_CA_BOLD,
	ANSI_CA_FAINT,
	ANSI_CA_ITALIC,
	ANSI_CA_UNDERLINE,
	ANSI_CA_BLINK,
	ANSI_CA_FASTBLINK,
	ANSI_CA_REVERSE,  /* or 'Inverse' technically.. .*/
	ANSI_CA_HIDDEN,  /* or 'Inverse' technically.. .*/
	ANSI_CA_CROSSEDOUT,
	ANSI_CA_DOUBLEUL,
	ANSI_CA_FRAMED,
	ANSI_CA_ENCIRCLED,
	ANSI_CA_OVERLINED,
	ANSI_CA_SUPERSCRIPT,
	ANSI_CA_SUBSCRIPT,
	ANSI_CA_MAX
} ansi_ca;

typedef struct {
	char *name;
	char *Ps[2];
} sgr_ca_t;

typedef enum {
	STYLE_FG,
	STYLE_BG,
	STYLE_ATTR,
	STYLE_MAX
} rgba_style_opt;

/* colorchip_t encapsulates a color, and a palette index */
typedef struct {
	char *name;
	char *definition;
	rgba_t color;
	int index;
} colorchip_t;


/* style_t encapsulates fgbg colors, and attributes. */
typedef struct {
	char *name;
	char *definition;
	rgba_t fg;
	rgba_t bg;
	gboolean attr[ANSI_CA_MAX];	/* array of enabled ca's */
	gboolean opt[STYLE_MAX];	/* array of enabled options */
} style_t;


/* common parameters */
typedef struct {
	double gamma;
	double contrast_ratio;  /* default of 1.0 means do nothing. */
	gboolean system_palette_fallback;  /* ok to search system palette */
	gboolean system_styles_fallback;  /* ok to search system styles */
	gboolean debug;
	colorchip_hash_t *palette;
	style_hash_t *styles;
	ansi_color_mode color_mode;
	ansi_fgbg fgbg_mode;
	rgba_t unmodified_fg; /* unmodified by min cr */
	style_t default_style; /* the default style */
	style_t current_style; /* currently on terminal */
} rgba_context;

/* function that returns a pointer to a new style based on a,b,amount, and
 * context. In general, if amount is 0, return a copy of a, if amount is 1.0
 * return a copy of b... but whatever.  Caller must free the returned style. */
typedef style_t *style_blendfn_t(
	style_t *a, 
	style_t *b, 
	double amount, 
	rgba_context *ctx
);

/* ---- Exported global variable declarations ---- */
extern sgr_ca_t ansi_attributes[];

/* ---- Exported function declarations ---- */

/* initialization routines */
void locolor_init(void);
void locolor_quit(void);
/* ... */

/* context manipulation */
ansi_fgbg color_mode_from_env(void);
rgba_context *new_rgba_context(void);
rgba_context *resolve_ctx(rgba_context *ctx);
void free_rgba_context(rgba_context *f);

/* type conversion */
rgba_t rgba_from_channel(channel_t r, channel_t g, channel_t b, channel_t a);
rgba_t rgba_from_double(double r, double g, double b, double a);
rgba_t rgba_from_drgba(drgba_t *d);
rgba_t rgba_from_hsva(hsva_t *h);
void rgba_to_ansi216idx(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a);
void rgba_to_channel(rgba_t color, channel_t *r, channel_t *g, channel_t *b, channel_t *a);
void rgba_to_double(rgba_t color, double *r, double *g, double *b, double *a);
void rgba_to_drgba(rgba_t color, drgba_t *d);
void rgba_to_hsva(rgba_t color, hsva_t *h);
/* ... */

/* color manipulation functions */
double rgba_contrast_ratio(rgba_t A, rgba_t B);
double rgba_relative_luminance(rgba_t color);
double unit_clip(double c);
int rgba_distance2(rgba_t c1, rgba_t c2);
rgba_t rgba_alphablend(rgba_t A, rgba_t B);
rgba_t rgba_alpha(rgba_t c,channel_t a);
rgba_t rgba_ansi216(rgba_t color);
rgba_t rgba_contrasting_color(rgba_t fg, rgba_t bg, double mincr);
rgba_t rgba_from_ansi256(int idx);
rgba_t rgba_gammablend(rgba_t A, rgba_t B, double y);
rgba_t rgba_gamma(rgba_t c,double y);
rgba_t rgba_greyscale(rgba_t color);
rgba_t rgba_hsvblend(rgba_t A, rgba_t B, double amount, int ccw);
rgba_t rgba_hsvblend_long(rgba_t A, rgba_t B, double amount);
rgba_t rgba_hsvblend_path(rgba_t A, rgba_t B, double amount, int longway);
rgba_t rgba_hsvblend_short(rgba_t A, rgba_t B, double amount);
rgba_t rgba_invert(rgba_t a);
rgba_t rgba_multiply(rgba_t a, rgba_t b);
rgba_t rgba_opaque(rgba_t c);
rgba_t rgba_posterize(rgba_t color, double N);
rgba_t rgba_screen(rgba_t a, rgba_t b);
/* ... */


/* ANSI Terminal code functions. */
int sgr_ansi216(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ansiAttr(char *dst, size_t dlen, ansi_ca attr, gboolean state, rgba_context *ctx);
int sgr_ansiColor(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ansiNormal(char *buf, size_t size, rgba_context *ctx);
int sgr_ansiStyle(char *dst, size_t dlen, style_t *style, rgba_context *ctx);
int sgr_ansiTruecolor(char *buf, size_t size, rgba_t color, rgba_context *ctx);
int sgr_ctxDefaults(char *buf, size_t size, rgba_context *ctx);
/* ... */

/* palette related functions */
int lookup_color_name(char *name, rgba_context *ctx, rgba_t *color);
int load_colorchip_hash(char *filename, rgba_context *ctx);
int rgba_sscanf(char *buf,rgba_t *c, rgba_context *ctx);
int colorchip_sscanf(char *buf,colorchip_t *c, rgba_context *ctx);
colorchip_t *new_colorchip(void);
colorchip_t *rgba_closest_match(rgba_t color, colorchip_t *pal, int from, int to);
colorchip_hash_t *new_palette(void);
rgba_t rgba_parse(char *name,rgba_context *ctx);
void free_colorchip(colorchip_t *f);
void free_colorchip_hash(colorchip_hash_t *f);

/* style related functions */
int load_style_hash(char *filename, rgba_context *ctx);
int style_sscanf(char *buf, style_t **styleloc, rgba_context *ctx);
style_hash_t *new_style_hash(void);
style_t *lookup_style_name(char *name, rgba_context *ctx);
style_t *new_style(void);
style_t *new_style_parse(char *name,rgba_context *ctx);
style_t *style_cpy(style_t *d, style_t *s);
style_t *style_dup(style_t *s);
void free_style(style_t *f);
void free_style_hash(style_hash_t *f);

/* fancy style blending and formatting functions */
int style_blend_text(char *dst,size_t dlen,char *src, GList *styles, style_blendfn_t *blendfn, rgba_context *ctx);
style_t *style_blendfn_generic( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_list(double amount,GList *slist, style_blendfn_t *blendfn,rgba_context *ctx);
style_t *style_blendfn_gamma( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_hsva( style_t *a, style_t *b, double amount, rgba_context *ctx);
style_t *style_blendfn_rainbow( style_t *a, style_t *b, double amount, rgba_context *ctx);
/* ... */


#endif /* LOCOLOR_H */
